""" This is the part where OJN stores BPM, Notes, Measure Changes and AutoPlays.

For each level (easy, normal, hard), the number of packages are specified.
So we know how many packages we need to loop through

For each package header, it defines the number of events that will occur after it.

By knowing the amount of events and the location of the note, we can find out the offset

Consider this::

            P E P E E E E P E E P E P E E E E E E E E
    Events  1   4         2     1   8
    Snap      1   1 2 3 4   1 2   1   1 2 3 4 5 6 7 8
    Measure 0   1         1     2   2
    Note      ^       ^       ^   ^               ^

From this, we can tell the offset of the notes by calculating the offset from a BPM List and taking a fraction of
the beat by looking at the snap.

Notice that some snaps have no data in them, these are useful to "pad" these notes into place if they have a complex
snap.

There are more specifications on other data decryption in open2jam_.

.. _open2jam: https://open2jam.wordpress.com/

"""

from __future__ import annotations

import logging
import struct
from collections import deque
from dataclasses import dataclass, field
from typing import List, Union, Dict

from reamber.o2jam.O2JBpmObj import O2JBpmObj
from reamber.o2jam.O2JHitObj import O2JHitObj
from reamber.o2jam.O2JHoldObj import O2JHoldObj

log = logging.getLogger(__name__)

class O2JConst:

    # These data are required to find out which notes are hits and holds
    HIT_BYTES      : bytes = b'\x00'
    HOLD_HEAD_BYTES: bytes = b'\x02'
    HOLD_TAIL_BYTES: bytes = b'\x03'
    # HIT            : int   = 0  # Not very useful
    # HOLD_HEAD      : int   = 2  # Not very useful
    # HOLD_TAIL      : int   = 3  # Not very useful


class O2JNoteChannel:
    MEASURE_FRACTION: int = 0
    BPM_CHANGE      : int = 1
    COL_1           : int = 2
    COL_2           : int = 3
    COL_3           : int = 4
    COL_4           : int = 5
    COL_5           : int = 6
    COL_6           : int = 7
    COL_7           : int = 8
    COL_RANGE       : range = range(2, 9)
    AUTOPLAY_1      : int = 9
    AUTOPLAY_2      : int = 10
    AUTOPLAY_3      : int = 11
    AUTOPLAY_4      : int = 12
    AUTOPLAY_5      : int = 13
    AUTOPLAY_6      : int = 14
    AUTOPLAY_7      : int = 15
    AUTOPLAY_8      : int = 16
    AUTOPLAY_9      : int = 17
    AUTOPLAY_10     : int = 18
    AUTOPLAY_11     : int = 19
    AUTOPLAY_12     : int = 20
    AUTOPLAY_13     : int = 21
    AUTOPLAY_14     : int = 22
    AUTOPLAY_RANGE  : range = range(9, 23)
    # Is there more?
    # Don't worry about the size of this obj, all of them are static.


@dataclass
class O2JEventMeasureChange:
    """ If the value is 0.75, the size of this measure will be only 75% of a normal measure. """
    # When the channel is 0(fractional measure), the 4 bytes are a float,
    # indicating how much of the measure is actually used,
    # so if the value is 0.75, the size of this measure will be only 75% of a normal measure.
    fracLength: float = 1.0

@dataclass
class O2JEventPackage:
    """ This class facilitates loading of Event Packages. """

    measure: int = 0  # Len 4 (Int)
    channel: int = -1  # Len 2 (Short)
    # There's a Len 2 (Short) indicating how many events there are.
    # Then it's followed by that amount of events.
    events: List[Union[O2JBpmObj, O2JHitObj, O2JHoldObj, O2JEventMeasureChange]] =\
        field(default_factory=lambda: [])

    @staticmethod
    def readEventPackages(data: bytes, lvlPkgCounts: List[int]) -> List[List[O2JEventPackage]]:
        """ Reads all events, this data found after the metadata (300:)

        :param lvlPkgCounts: The count of pkgs per level
        :param data: All the event data in bytes.
        """

        # Don't think we can reliably get all the offsets of notes, we'll firstly find their measures, then calculate
        # the offsets separately.
        # I have doubts because of bpm changes

        # Because of this, we have to parse LNs in a different wave instead of dynamically like stepmania.

        # Additional variables
        # O2JHit.measure
        # O2JHold.measure
        # O2JHold.tailMeasure

        lvls: List[List[O2JEventPackage]] = []
        dataQ = deque(data)

        # Column, Offset
        holdBuffer: Dict[int, O2JHoldObj] = {}

        # For each level, we will read the required amount of packages, then go to the next
        for lvlPkgI, lvlPkgCount in enumerate(lvlPkgCounts):
            log.debug(f"Loading New Level with {lvlPkgCount} Packages")
            # noinspection PyTypeChecker
            lvlPkg: List[O2JEventPackage] = [None] * lvlPkgCount
            for pkgI in range(0, lvlPkgCount):
                if len(dataQ) == 0: break
                pkg = O2JEventPackage()
                pkgData = []
                for i in range(8): pkgData.append(dataQ.popleft())

                pkg.measure = struct.unpack("<i", bytes(pkgData[0:4]))[0]
                pkg.channel = struct.unpack("<h", bytes(pkgData[4:6]))[0]
                eventCount  = struct.unpack("<h", bytes(pkgData[6:8]))[0]

                eventsData = []
                for i in range(4 * eventCount): eventsData.append(dataQ.popleft())
                eventsData = bytes(eventsData)

                if pkg.channel in O2JNoteChannel.COL_RANGE:
                    pkg.events += O2JEventPackage.readEventsNote(eventsData,
                                                                 pkg.channel - 2,  # Package CHN - 2 is column
                                                                 holdBuffer,
                                                                 pkg.measure)
                elif pkg.channel == O2JNoteChannel.BPM_CHANGE:
                    pkg.events += O2JEventPackage.readEventsBpm(eventsData,
                                                                pkg.measure)
                elif pkg.channel == O2JNoteChannel.MEASURE_FRACTION:
                    measureFrac = O2JEventPackage.readEventsMeasure(eventsData)
                    pkg.events.append(O2JEventMeasureChange(measureFrac))
                else:
                    # log.debug(f"{currMeasure} count: {eventCount}, chn: {pkg.channel}")
                    pass
                lvlPkg[pkgI] = pkg
            lvls.append(lvlPkg)
        return lvls

    @staticmethod
    def readEventsMeasure(eventsData: bytes) -> float:
        """ Reads the fractional measure data.

        This may not work as intended as there is no ojn files to test this feature.

        :param eventsData: The 4 byte data point to unpack.
        :return: Returns a float indicating the fraction of the measure.
        """
        log.warning("Fractional measure isn't confidently implemented, conversion may fail.")
        return struct.unpack("<f", eventsData[0:4])[0]

    @staticmethod
    def readEventsBpm(eventsData: bytes, currMeasure: float) -> List[O2JBpmObj]:
        """ Reads the event's bpms.

        Just like the Notes, this can have disabled points where Bpm == 0, that means there's no bpm there.

        This is under the presumption that Bpm == 0 is not supported at all.

        :param eventsData: The bytes to unpack.
        :param currMeasure: The current measure, used to calculate the offset later.
        :return: Returns a List of Bpm Points found.
        """
        eventCount = int(len(eventsData) / 4)
        bpms = []

        for i in range(eventCount):
            bpm = struct.unpack("<f", eventsData[i*4:(i+1)*4])[0]
            if bpm == 0: continue
            log.debug(f"Appended BPM {bpm} at {currMeasure + i / eventCount}")
            bpmObj = O2JBpmObj(bpm=bpm)
            bpmObj.measure = currMeasure + i / eventCount
            bpms.append(bpmObj)

        return bpms

    @staticmethod
    def readEventsNote(eventsData: bytes, column: int, holdBuffer: Dict[int, O2JHoldObj], currMeasure: float) ->\
            List[Union[O2JHitObj, O2JHoldObj]]:
        """ Reads the event's notes.

        This can have disabled points dictated by the first 2 bytes (see: 'enabled')

        :param eventsData: The bytes to unpack.
        :param column: The current column, this can be found in the header.
        :param holdBuffer: The hold buffer, this acts like a static variable to facilitate head and tail matching.
        :param currMeasure: The current measure, used to calculate the offset later.
        :return: Returns a List of Bpm Points found.
        """

        notes = []

        eventCount = int(len(eventsData) / 4)

        for i in range(eventCount):
            enabled  = struct.unpack("<h", bytes(eventsData[0 + i * 4:2 + i * 4]))[0]
            if enabled == 0: continue

            subMeasure = i / eventCount + currMeasure
            volumePan  = struct.unpack("<s", eventsData[2 + i * 4:3 + i * 4])[0]
            volume     = int.from_bytes(volumePan, 'little') // 16
            pan        = int.from_bytes(volumePan, 'little') % 16
            noteType   = struct.unpack("<c", eventsData[3 + i * 4:4 + i * 4])[0]
            log.debug(f"Event Data: {eventsData[0 + i * 4:4 + i * 4]}")

            if noteType == O2JConst.HIT_BYTES:
                hit = O2JHitObj(volume=volume, pan=pan, offset=0, column=column)
                hit.measure = subMeasure
                notes.append(hit)
                log.debug(f"Appended Note {column} at {subMeasure}")
            elif noteType == O2JConst.HOLD_HEAD_BYTES:
                hold = O2JHoldObj(volume=volume, pan=pan, column=column, length=-1)
                hold.measure = subMeasure
                holdBuffer[column] = hold
            elif noteType == O2JConst.HOLD_TAIL_BYTES:
                hold = holdBuffer.pop(column)
                hold.tailMeasure = subMeasure
                notes.append(hold)
                log.debug(f"Appended LNote {column} at {subMeasure}, Tail:{hold.tailMeasure}")
            else:
                pass
        return notes



from pandas import DataFrame
from typing import List, Dict

from reamber.base.MapObj import MapObj
from dataclasses import asdict


def asDataFrame(obj: MapObj) -> Dict[str, DataFrame]:
    """ Converts a MapObj to a Dictionary of DataFrames

    :param obj: Any MapObj
    :return: A Dictionary of DataFrames
    """
    out: Dict = {}
    for key, item in obj.__dict__.items():
        if not isinstance(item, List): continue  # Don't create DF for singles
        try:
            out[key] = DataFrame([asdict(o) for o in item])
        except TypeError:  # asdict would throw this
            continue
        except Exception as e:  # This should be some other error
            print(e)
    return out







from reamber.base.lists.notes.NoteList import NoteList
from reamber.base.NoteObj import NoteObj
from typing import List
from abc import ABC


class SMNoteList(NoteList, ABC):
    def data(self) -> List[NoteObj]: pass




from reamber.base.lists.NotePkg import NotePkg
from typing import overload
from reamber.algorithms.analysis.bpm.aveBpm import aveBpm
from reamber.algorithms.analysis.generic.rollingDensity import rollingDensity
from reamber.algorithms.analysis.describe.meta import mapMetadata

from reamber.o2jam.O2JMapSetObj import O2JMapSetObj, O2JMapObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.quaver.QuaMapObj import QuaMapObj

from reamber.o2jam.lists.O2JNotePkg import O2JNotePkg
from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.sm.lists.SMNotePkg import SMNotePkg
from reamber.quaver.lists.QuaNotePkg import QuaNotePkg
import datetime


@overload
def describe(m: O2JMapObj, s: O2JMapSetObj = None) -> None: ...
@overload
def describe(m: OsuMapObj, s: None = None) -> None: ...
@overload
def describe(m: QuaMapObj, s: None = None) -> None: ...
@overload
def describe(m: SMMapObj, s: SMMapSetObj = None) -> None: ...
def describe(m: QuaMapObj, s=None, rounding: int = 2, unicode: bool = False) -> None:
    """ Describes the map's attributes as a short summary

    :param m: The map
    :param s: The mapset, may be optional for some types
    :param rounding: The decimal rounding
    :param unicode: Whether to attempt to get the non-unicode or unicode. \
        Doesn't attempt to translate.
    """
    print(f"Average BPM: {round(aveBpm(m), rounding)}")

    first, last = m.notes.firstLastOffset()
    print(f"Map Length: {datetime.timedelta(milliseconds=last - first)}")
    # noinspection PyTypeChecker
    print(mapMetadata(m=m, s=s, unicode=unicode))
    print("---- NPS ----")
    print("All:", end='  ')
    describeNotes(m.notes)
    for key in range(m.notes.maxColumn() + 1):
        print(f"Col{key}:", end=' ')
        describeNotes(m.notes.inColumns([key], inplace=False))
    pass

@overload
def describeNotes(notePkg: OsuNotePkg, rounding: int = 2): ...
@overload
def describeNotes(notePkg: SMNotePkg, rounding: int = 2): ...
@overload
def describeNotes(notePkg: QuaNotePkg, rounding: int = 2): ...
@overload
def describeNotes(notePkg: O2JNotePkg, rounding: int = 2): ...
def describeNotes(notePkg: NotePkg, rounding: int = 2):
    """ Describes a single NotePkg

    Prints out Count, Median, 75% quantile and max

    :param notePkg: Any Note Package
    :param rounding: The decimal rounding
    """
    # This is fixed to be 1 for consistency in value
    sr = rollingDensity([i for j in notePkg.offsets().values() for i in j], rollingWindowS=1)
    print(       f"Count: {len([i for j in notePkg.offsets().values() for i in j])}, "
          f"50% (Median): {float(sr.quantile(0.5)):.{rounding}f}, "
                   f"75%: {float(sr.quantile(0.75)):.{rounding}f}, "
            f"100% (Max): {float(sr.max()):.{rounding}f}")




import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence
import logging

logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)

class TestNormalize(unittest.TestCase):

    def testGeneric(self):
        seq = SvSequence([0, (500, 1.5, True), 1000])
        seq.normalizeTo(inplace=True)
        self.assertAlmostEqual(seq[0].multiplier, 0.5)

    def testComplex(self):
        seq = SvSequence([0, (500, 1.5, True), (700, 0.75, True), (800, 1.5, False), 1000])
        seq.normalizeTo(inplace=True)
        self.assertAlmostEqual(seq[0].multiplier, 0.78125)
        self.assertAlmostEqual(seq[3].multiplier, 1.171875)

    def testIgnoreFixed(self):
        seq = SvSequence([0, (500, 1.5, True), 1000])
        seq.normalizeTo(inplace=True, ignoreFixed=True)
        self.assertAlmostEqual(seq[0].multiplier, 0.8)
        self.assertAlmostEqual(seq[1].multiplier, 1.2)

    def testBadMin(self):
        seq = SvSequence([0, (500, 2.1, True), 1000])
        self.assertRaises(AssertionError, seq.normalizeTo, inplace=True, minAllowable=0)

    def testBadMax(self):
        # Norm Test
        seq = SvSequence([0, (500, 0.5, True), 1000])
        self.assertRaises(AssertionError, seq.normalizeTo, inplace=True, maxAllowable=1.2)


if __name__ == '__main__':
    unittest.main()



from reamber.osu.OsuSampleSet import OsuSampleSet
from dataclasses import dataclass
from math import ceil


class OsuNoteObjType:
    NOTE: int = 1
    LONG_NOTE: int = 128


@dataclass
class OsuNoteObjMeta:
    """ Holds all metadata for every note object"""

    hitsoundSet: int = OsuSampleSet.AUTO
    sampleSet: int = OsuSampleSet.AUTO
    additionSet: int = OsuSampleSet.AUTO
    customSet: int = OsuSampleSet.AUTO
    volume: int = 0
    hitsoundFile: str = ""
    # keys to be supplied by map

    @staticmethod
    def xAxisToColumn(xAxis: float, keys: int, clip: bool = True) -> int:
        """ Converts the xAxis code in .osu to an actual column value

        Note that column starts from 0

        :param xAxis: The code in .osu to convert
        :param keys: Required for conversion
        :param clip: If true the return will be clipped to max of (keys - 1)
        :return: The actual column value, starting from 0
        """
        col = int(ceil((xAxis * keys - 256.0) / 512.0))
        return min(keys - 1, col) if clip else col

    @staticmethod
    def columnToXAxis(column: float, keys: int) -> int:
        """ Converts the actual column value to a .osu writable code value

        Note that column starts from 0

        :param column: The column to convert
        :param keys: Required for conversion
        :return: The actual code
        """
        return int(round(((512.0 * column) + 256.0) / keys))

    @staticmethod
    def isHitObj(s: str):
        """ Checks if the string is a HitObject """
        return s.count(":") == 4

    @staticmethod
    def isHoldObj(s: str):
        """ Checks if the string is a HoldObject """
        return s.count(":") == 5



from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.quaver.QuaHitObj import QuaHitObj
from reamber.quaver.QuaHoldObj import QuaHoldObj
from reamber.quaver.QuaBpmObj import QuaBpmObj
from reamber.base.BpmObj import BpmObj
from reamber.quaver.lists.QuaNotePkg import QuaNotePkg
from reamber.quaver.lists.QuaBpmList import QuaBpmList
from reamber.quaver.lists.notes.QuaHoldList import QuaHoldList
from reamber.quaver.lists.notes.QuaHitList import QuaHitList
from typing import List


class SMToQua:
    @staticmethod
    def convert(sm: SMMapSetObj) -> List[QuaMapObj]:
        """ Converts a SMMapset to possibly multiple quaver maps

        Note that a mapset contains maps, so a list would be expected.
        SMMap conversion is not possible due to lack of SMMapset Metadata

        :param sm: SM Mapset
        :return: List of Quaver Maps
        """
        quaMapSet: List[QuaMapObj] = []
        for smMap in sm.maps:
            assert isinstance(smMap, SMMapObj)
            hits: List[QuaHitObj] = []
            holds: List[QuaHoldObj] = []

            # Note Conversion
            for hit in smMap.notes.hits():
                hits.append(QuaHitObj(offset=hit.offset, column=hit.column))
            for hold in smMap.notes.holds():
                holds.append(QuaHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

            bpms: List[BpmObj] = []

            # Timing Point Conversion
            for bpm in smMap.bpms:
                bpms.append(QuaBpmObj(offset=bpm.offset, bpm=bpm.bpm))

            # Extract Metadata
            quaMap = QuaMapObj(
                backgroundFile=sm.background,
                title=sm.title,
                artist=sm.artist,
                audioFile=sm.music,
                creator=sm.credit,
                difficultyName=f"{smMap.difficulty} {smMap.difficultyVal}",
                songPreviewTime=int(sm.sampleStart),
                bpms=QuaBpmList(bpms),
                notes=QuaNotePkg(hits=QuaHitList(hits),
                                 holds=QuaHoldList(holds))
            )
            quaMapSet.append(quaMap)
        return quaMapSet



from dataclasses import dataclass
from reamber.base.TimedObj import TimedObj


@dataclass
class SvObj(TimedObj):
    """ This isn't a base object since it's not a required object for any map.

    This is just to facilitate SvSequence and SvPkg

    Offset uses milliseconds units. Negative values are allowed

    A Fixed Sv means recalculation will avoid changing the value unless explicitly stated.
    """

    multiplier: float = 1.0
    fixed: bool = False



from reamber.base.RAConst import RAConst
from typing import Callable, List, Union
from reamber.algorithms.generate.sv.generators.svFuncSequencer import svFuncSequencer
from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.osu.OsuSvObj import OsuSvObj,MAX_SV,MIN_SV

from copy import deepcopy

def svOsuMeasureLineB(firstOffset: float,
                      lastOffset: float,
                      funcs: List[Callable[[float], float]],
                      referenceBpm: float,
                      endBpm: float,
                      paddingSize: int = 10,
                      stopBpm: float = 1e-05,
                      fillBpm: float or None = 1e07,
                      startX: float = 0,
                      endX: float = 1) -> List[Union[OsuSvObj, OsuBpmObj]]:
    """ Generates Measure Line movement for osu! maps. Version 2. Inspired by datoujia

    This one directly returns svs and bpms due to the nature of the algorithm requiring osu! objects.

    This could output as Quaver but it has negative scroll

    ``S{_}...D{F},S{_}...D{F}_,...``

    :param firstOffset: The first Offset to start the function (x = startX)
    :param lastOffset: The last Offset to end the function (x = endX)
    :param funcs: The functions to use. startX <= x <= endX will be called, expecting a BPM as an output. \
        The more functions you have, the "laggier" it will be.
    :param referenceBpm: The bpm that is used to zero. Found by looking at BPM:XXX-XXX(Reference Bpm) in song select.
    :param endBpm: The bpm value referenced for Bpms.
    :param paddingSize: The size of the padding, the larger the value, the lower the FPS
    :param stopBpm: The bpm value for stop Bpms. Cannot be 0.
    :param fillBpm: The bpm to use to fill such that the sequence ends on lastOffset. None for no fill.
    :param startX: The starting X to use
    :param endX: The ending X to use
    """

    # Optimized value to make sure that 1.0 in input means at the top of the screen.
    # Not accurate for all scrolls and different hit positions.
    SCALING_FACTOR = 310.375 / referenceBpm

    depBpm = 60000 * (len(funcs) + 1)

    totalGaps = RAConst.mSecToMin(depBpm)  # The number of measure lines/gaps generated.

    # Append a y = 0 to get diff on first func
    funcs_ = [lambda x: 0, *funcs]
    funcDiff = []
    for funcI in range(len(funcs)):  # -1 due to the appended y = 0, -1 due to custom last func
        def f(x, i=funcI):
            sort = sorted([g(x) * SCALING_FACTOR  for g in funcs_])
            for s in range(len(sort)):
                sort[s] = max(0.0, sort[s])  # We eliminate all negative inputs

            diff = [g2 - g1 for g1, g2 in zip(sort[:-1], sort[1:])]

            for d in range(len(diff)):
                if diff[d] < MIN_SV:
                    if d < len(diff) - 1:
                        diff[d + 1] -= MIN_SV - diff[d]
                    diff[d] = MIN_SV

            return diff[i]
        funcDiff.append(deepcopy(f))

    repeats = int((lastOffset - firstOffset) / (paddingSize + 3))

    bpmPkg = svFuncSequencer([stopBpm,
                              *[None for _ in range(paddingSize)],
                              depBpm,
                              None],
                             offsets=1,
                             repeatGap=2,
                             repeats=repeats).addOffset(firstOffset, inplace=False)

    svPkg = svFuncSequencer([*funcDiff, MAX_SV],
                            offsets=1 / totalGaps,
                            repeats=repeats,
                            repeatGap=2 + paddingSize + (1 - len(funcs) / totalGaps),
                            startX=startX,
                            endX=endX).addOffset(by=1 + paddingSize + firstOffset, inplace=False)

    # We clip the values here, just to optimize the output a bit
    svOsu = svPkg.combine().writeAsSv(OsuSvObj)
    for sv in svOsu:
        assert isinstance(sv, OsuSvObj)
        sv.multiplier = min(MAX_SV, sv.multiplier)
        sv.multiplier = max(MIN_SV, sv.multiplier)

    # Combines both sequence together and writes them as osu
    outList = [*bpmPkg.combine().writeAsBpm(OsuBpmObj, metronome=1), *svOsu]

    if fillBpm is not None:
        outList.extend([*[OsuBpmObj(x, fillBpm) for x in range(int(firstOffset + (3 + paddingSize) * repeats),
                                                               int(lastOffset))],
                        OsuBpmObj(lastOffset, endBpm)])
    else:
        outList.append(OsuBpmObj(int(firstOffset + (3 + paddingSize) * repeats), endBpm))

    return sorted(outList, key=lambda x: x.offset)



from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.quaver.QuaMapObjMeta import QuaMapObjMode
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.osu.OsuHitObj import OsuHitObj
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.osu.OsuSvObj import OsuSvObj
from reamber.base.BpmObj import BpmObj

from reamber.osu.lists.OsuBpmList import OsuBpmList
from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.osu.lists.notes.OsuHoldList import OsuHoldList
from reamber.osu.lists.notes.OsuHitList import OsuHitList
from reamber.osu.lists.OsuSvList import OsuSvList
from typing import List


class QuaToOsu:
    @staticmethod
    def convert(qua: QuaMapObj) -> OsuMapObj:
        """ Converts a Quaver map to an osu map

        :param qua: Quaver map
        :return: Osu Map
        """

        hits: List[OsuHitObj] = []
        holds: List[OsuHoldObj] = []

        # Note Conversion
        for hit in qua.notes.hits():
            hits.append(OsuHitObj(offset=hit.offset, column=hit.column))
        for hold in qua.notes.holds():
            holds.append(OsuHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

        bpms: List[BpmObj] = []
        svs: List[OsuSvObj] = []
        # Timing Point Conversion
        for bpm in qua.bpms:
            bpms.append(OsuBpmObj(offset=bpm.offset, bpm=bpm.bpm))

        for sv in qua.svs:
            svs.append(OsuSvObj(offset=sv.offset, multiplier=sv.multiplier))

        # Extract Metadata
        osuMap = OsuMapObj(
            backgroundFileName=qua.backgroundFile,
            title=qua.title,
            circleSize=QuaMapObjMode.keys(qua.mode),
            titleUnicode=qua.title,
            artist=qua.artist,
            artistUnicode=qua.artist,
            audioFileName=qua.audioFile,
            creator=qua.creator,
            version=qua.difficultyName,
            previewTime=qua.songPreviewTime,
            bpms=OsuBpmList(bpms),
            svs=OsuSvList(svs),
            notes=OsuNotePkg(hits=OsuHitList(hits),
                             holds=OsuHoldList(holds))
        )

        return osuMap



from reamber.sm.SMBpmObj import SMBpmObj
from reamber.sm.SMStopObj import SMStopObj

from reamber.base.BpmObj import BpmObj

from reamber.base.RAConst import RAConst

from dataclasses import dataclass
from dataclasses import field
from typing import List


@dataclass
class SMMapSetObjMeta:
    title: str = ""
    subtitle: str = ""
    artist: str = ""
    titleTranslit: str = ""
    subtitleTranslit: str = ""
    artistTranslit: str = ""
    genre: str = ""
    credit: str = ""
    banner: str = ""
    background: str = ""
    lyricsPath: str = ""
    cdTitle: str = ""
    music: str = ""
    offset: float = None  # Offset is None as we do a comparison on offset, see SMMapSetObj.py::_readBpms
    _bpmsStr: List[str] = field(default_factory=lambda: [])
    _stopsStr: List[str] = field(default_factory=lambda: [])
    stops: List[SMStopObj] = field(default_factory=lambda: [])
    sampleStart: float = 0.0
    sampleLength: float = 10.0
    displayBpm: str = ""
    selectable: bool = True
    bgChanges: str = ""  # Idk what this does
    fgChanges: str = ""  # Idk what this does

    def _readMetadata(self, lines: List[str]):
        for line in lines:
            if line == "":
                continue

            s = [token.strip() for token in line.split(":")]
            # This is to get rid of comments
            # e.g.
            # // HELLO\n#TITLE:WORLD -> #TITLE:WORLD
            if len(s[0]) == 0: continue
            if not s[0].startswith("#"): s[0] = s[0][s[0].rfind('#'):]

            if   s[0] == "#TITLE":              self.title = s[1].strip()
            elif s[0] == "#SUBTITLE":           self.subtitle = s[1].strip()
            elif s[0] == "#ARTIST":             self.artist = s[1].strip()
            elif s[0] == "#TITLETRANSLIT":      self.titleTranslit = s[1].strip()
            elif s[0] == "#SUBTITLETRANSLIT":   self.subtitleTranslit = s[1].strip()
            elif s[0] == "#ARTISTTRANSLIT":     self.artistTranslit = s[1].strip()
            elif s[0] == "#GENRE":              self.genre = s[1].strip()
            elif s[0] == "#CREDIT":             self.credit = s[1].strip()
            elif s[0] == "#BANNER":             self.banner = s[1].strip()
            elif s[0] == "#BACKGROUND":         self.background = s[1].strip()
            elif s[0] == "#LYRICSPATH":         self.lyricsPath = s[1].strip()
            elif s[0] == "#CDTITLE":            self.cdTitle = s[1].strip()
            elif s[0] == "#MUSIC":              self.music = s[1].strip()
            elif s[0] == "#OFFSET":             self.offset = RAConst.secToMSec(float(s[1].strip()))
            elif s[0] == "#BPMS":               self._bpmsStr = s[1].strip().split(",")
            elif s[0] == "#STOPS":              self._stopsStr = s[1].strip().split(",")
            elif s[0] == "#SAMPLESTART":        self.sampleStart = RAConst.secToMSec(float(s[1].strip()))
            elif s[0] == "#SAMPLELENGTH":       self.sampleLength = RAConst.secToMSec(float(s[1].strip()))
            elif s[0] == "#DISPLAYBpm":         self.displayBpm = s[1].strip()
            elif s[0] == "#SELECTABLE":         self.selectable = True if s[1].strip() == "YES" else False
            elif s[0] == "#BGCHANGES":          self.bgChanges = s[1].strip()
            elif s[0] == "#FGCHANGES":          self.fgChanges = s[1].strip()

    def _writeMetadata(self, bpms: List[BpmObj]) -> List[str]:
        bpms.sort(key=lambda tp: tp.offset)

        bpmBeats = SMBpmObj.getBeats(bpms, bpms)
        stopBeats = SMBpmObj.getBeats(self.stops, bpms)

        return [
            f"#TITLE:{self.title};",
            f"#SUBTITLE:{self.subtitle};",
            f"#ARTIST:{self.artist};",
            f"#TITLETRANSLIT:{self.titleTranslit};",
            f"#SUBTITLETRANSLIT:{self.subtitleTranslit};",
            f"#ARTISTTRANSLIT:{self.artistTranslit};",
            f"#GENRE:{self.genre};",
            f"#CREDIT:{self.credit};",
            f"#BANNER:{self.banner};",
            f"#BACKGROUND:{self.background};",
            f"#LYRICSPATH:{self.lyricsPath};",
            f"#CDTITLE:{self.cdTitle};",
            f"#MUSIC:{self.music};",
            f"#OFFSET:{RAConst.mSecToSec(self.offset)};",
            f"#BPMS:" + ",\n".join([f"{beat}={bpm.bpm}" for beat, bpm in zip(bpmBeats, bpms)]) + ";",
            f"#STOPS:" + ",\n".join([f"{beat}={RAConst.mSecToSec(stop.length)}" for
                                     beat, stop in zip(stopBeats, self.stops)]) + ";",
            f"#SAMPLESTART:{RAConst.mSecToSec(self.sampleStart)};",
            f"#SAMPLELENGTH:{RAConst.mSecToSec(self.sampleLength)};",
            f"#DISPLAYBpm:{self.displayBpm};",
            f"#SELECTABLE:" + "YES;" if self.selectable else "NO;",
            f"#BGCHANGES:{self.bgChanges};",
            f"#FGCHANGES:{self.fgChanges};",
        ]



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence


class TestRescale(unittest.TestCase):

    def test(self):
        # Rescale Test
        seq = SvSequence([0, 1, 2])
        seq.rescale(0, 1000, inplace=True)

        self.assertAlmostEqual(seq[0].offset, 0)
        self.assertAlmostEqual(seq[1].offset, 500)
        self.assertAlmostEqual(seq[2].offset, 1000)


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass, field
from typing import List, Dict


class QuaMapObjMode:
    KEYS_4: str = "Keys4"
    KEYS_7: str = "Keys7"

    @staticmethod
    def keys(s: str) -> int:
        """ Gets the keys as integer instead of string """
        if   s == QuaMapObjMode.KEYS_4: return 4
        elif s == QuaMapObjMode.KEYS_7: return 7
        else: return -1

    @staticmethod
    def str(i: int) -> str:
        """ Gets the keys as string instead of int """
        if   i == 4: return QuaMapObjMode.KEYS_4
        elif i == 7: return QuaMapObjMode.KEYS_7
        else: return ""


@dataclass
class QuaMapObjMeta:
    audioFile: str                  = ""
    songPreviewTime: int            = 0
    backgroundFile: str             = ""
    mapId: int                      = -1
    mapSetId: int                   = -1
    mode: str                       = QuaMapObjMode.KEYS_4
    title: str                      = ""
    artist: str                     = ""
    source: str                     = ""
    tags: str                       = ""  # Stated as '', should it be []? List[str]?
    creator: str                    = ""
    difficultyName: str             = ""
    description: str                = ""
    editorLayers: List[str]         = field(default_factory=lambda: [])
    customAudioSamples: List[str]   = field(default_factory=lambda: [])
    soundEffects: List[str]         = field(default_factory=lambda: [])

    def _readMetadata(self, d: Dict):
        """ Reads the Metadata Dictionary provided by the YAML Library
        :param d: This is simply the whole Dictionary after the TPs and Notes are popped
        """
        self.audioFile          = d['AudioFile']
        self.songPreviewTime    = d['SongPreviewTime']
        self.backgroundFile     = d['BackgroundFile']
        self.mapId              = d['MapId']
        self.mapSetId           = d['MapSetId']
        self.mode               = d['Mode']
        self.title              = d['Title']
        self.artist             = d['Artist']
        self.source             = d['Source']
        self.tags               = d['Tags']
        self.creator            = d['Creator']
        self.difficultyName     = d['DifficultyName']
        self.description        = d['Description']
        self.editorLayers       = d['EditorLayers']
        self.customAudioSamples = d['CustomAudioSamples']
        self.soundEffects       = d['SoundEffects']

    def _writeMeta(self) -> Dict:
        """ Writes the metadata as a Dictionary and returns it """
        return {
            'AudioFile': self.audioFile,
            'SongPreviewTime': self.songPreviewTime,
            'BackgroundFile': self.backgroundFile,
            'MapId': self.mapId,
            'MapSetId': self.mapSetId,
            'Mode': self.mode,
            'Title': self.title,
            'Artist': self.artist,
            'Source': self.source,
            'Tags': self.tags,
            'Creator': self.creator,
            'DifficultyName': self.difficultyName,
            'Description': self.description,
            'EditorLayers': self.editorLayers,
            'CustomAudioSamples': self.customAudioSamples,
            'SoundEffects': self.soundEffects
        }



from reamber.osu.OsuMapObjMeta import OsuMapObjMeta
from reamber.base.MapObj import MapObj
from reamber.osu.lists.OsuSvList import OsuSvList

from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.osu.OsuSvObj import OsuSvObj
from reamber.osu.OsuHitObj import OsuHitObj
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.osu.OsuNoteObjMeta import OsuNoteObjMeta

from typing import List, Dict
from dataclasses import dataclass, field

from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.osu.lists.OsuBpmList import OsuBpmList
from reamber.osu.OsuSampleSet import OsuSampleSet
from reamber.base.lists.TimedList import TimedList

@dataclass
class OsuMapObj(MapObj, OsuMapObjMeta):

    notes: OsuNotePkg = field(default_factory=lambda: OsuNotePkg())
    bpms:  OsuBpmList = field(default_factory=lambda: OsuBpmList())
    svs:   OsuSvList  = field(default_factory=lambda: OsuSvList())

    def data(self) -> Dict[str, TimedList]:
        """ Gets the notes, bpms and svs as a dictionary """
        return {'notes': self.notes,
                'bpms': self.bpms,
                'svs': self.svs}

    def resetAllSamples(self, notes=True, samples=True) -> None:
        """ Resets all hitsounds and samples

        :param notes: Whether to reset hitsounds on notes
        :param samples: Whether to reset samples
        """
        if notes:
            for n in self.notes.hits():
                n.hitsoundFile = ""
                n.sampleSet = OsuSampleSet.AUTO
                n.hitsoundSet = OsuSampleSet.AUTO
                n.customSet = OsuSampleSet.AUTO
                n.additionSet = OsuSampleSet.AUTO

            for n in self.notes.holds():
                n.hitsoundFile = ""
                n.sampleSet = OsuSampleSet.AUTO
                n.hitsoundSet = OsuSampleSet.AUTO
                n.customSet = OsuSampleSet.AUTO
                n.additionSet = OsuSampleSet.AUTO

        if samples: self.samples.clear()

    def readFile(self, filePath=""):
        """ Reads a .osu, loads inplace, hence it doesn't return anything

        :param filePath: The path to the .osu file."""

        self.__init__()

        with open(filePath, "r", encoding="utf8") as f:
            file = f.read()
            file = file.replace("[TimingPoints]\n", "[HitObjects]\n")  # This is so as to split multiple delimiters
            fileSpl = file.split("[HitObjects]\n")
            if len(fileSpl) != 3:
                raise Exception("Incorrect File Format")

            self._readFileMetadata(fileSpl[0].split("\n"))

            for line in fileSpl[1].split("\n"):
                self._readFileTimingPoints(line)

            for line in fileSpl[2].split("\n"):
                self._readFileHitObjects(line)

    def writeFile(self, filePath=""):
        """ Writes a .osu, doesn't return anything.

        :param filePath: The path to a new .osu file."""

        with open(filePath, "w+", encoding="utf8") as f:
            for s in self.writeStringList():
                f.write(s + "\n")

            f.write("\n[TimingPoints]\n")
            for tp in self.bpms:
                assert isinstance(tp, OsuBpmObj)
                f.write(tp.writeString() + "\n")

            for tp in self.svs:
                assert isinstance(tp, OsuSvObj)
                f.write(tp.writeString() + "\n")

            f.write("\n[HitObjects]\n")
            for ho in self.notes.hits():
                f.write(ho.writeString(keys=int(self.circleSize)) + "\n")

            for ho in self.notes.holds():
                f.write(ho.writeString(keys=int(self.circleSize)) + "\n")

    def _readFileMetadata(self, lines: List[str]):
        """ Reads the metadata only, inclusive of Events """
        self.readStringList(lines)

    def _readFileTimingPoints(self, line: str):
        """ Reads all TimingPoints """
        if OsuTimingPointMeta.isSliderVelocity(line):
            self.svs.append(OsuSvObj.readString(line))
        elif OsuTimingPointMeta.isTimingPoint(line):
            self.bpms.append(OsuBpmObj.readString(line))

    def _readFileHitObjects(self, line: str):
        """ Reads all HitObjects """
        if OsuNoteObjMeta.isHitObj(line):
            self.notes.hits().append(OsuHitObj.readString(line, int(self.circleSize)))
        elif OsuNoteObjMeta.isHoldObj(line):
            self.notes.holds().append(OsuHoldObj.readString(line, int(self.circleSize)))



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.OsuToQua import OsuToQua
from reamber.osu.OsuMapObj import OsuMapObj
# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestOsuToQua(unittest.TestCase):

    # @profile
    def test_osu1(self):
        # Complex BPM Points
        osu = OsuMapObj()
        osu.readFile(OSU_CARAVAN)

        qua = OsuToQua.convert(osu)
        qua.writeFile("out.qua")

    # @profile
    def test_osu2(self):
        # Stops
        osu = OsuMapObj()
        osu.readFile(OSU_ESCAPES)

        qua = OsuToQua.convert(osu)
        qua.writeFile("out.qua")

    # @profile
    def test_osu3(self):
        # Complex BPM
        osu = OsuMapObj()
        osu.readFile(OSU_GRAVITY)

        qua = OsuToQua.convert(osu)
        qua.music = "Gravity.mp3"
        qua.writeFile("out.qua")


if __name__ == '__main__':
    unittest.main()



from reamber.algorithms.generate.sv.SvSequence import SvSequence
from reamber.base.lists.BpmList import BpmList


def svNormalizeBpm(bpms: BpmList,
                   normalizeTo: float) -> SvSequence:
    """ Generates Normalizing SVs for Bpm.

    Useful if Bpms are unexpectedly changing scroll speeds. Generated Scroll Velocities can counter it.

    :param bpms: Any Type of BpmList
    :param normalizeTo: The Bpm to normalize to
    """
    return SvSequence([(bpm.offset, normalizeTo / bpm.bpm) for bpm in bpms])



from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.quaver.QuaMapObjMeta import QuaMapObjMeta
from reamber.quaver.QuaSvObj import QuaSvObj
from reamber.quaver.QuaBpmObj import QuaBpmObj
from reamber.quaver.QuaHoldObj import QuaHoldObj
from reamber.quaver.QuaNoteObjMeta import QuaNoteObjMeta
from reamber.quaver.QuaHitObj import QuaHitObj


__all__ = ['QuaMapObj', 'QuaNoteObjMeta', 'QuaSvObj', 'QuaBpmObj', 'QuaHoldObj', 'QuaMapObjMeta',
           'QuaHitObj']



from __future__ import annotations
from reamber.base.lists.TimedList import TimedList
from reamber.quaver.QuaSvObj import QuaSvObj
from typing import List


class QuaSvList(List[QuaSvObj], TimedList):

    def _upcast(self, objList: List = None) -> QuaSvList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: QuaSvList
        """
        return QuaSvList(objList)

    def data(self) -> List[QuaSvObj]:
        return self

    def multipliers(self) -> List[float]:
        return self.attribute('multiplier')



from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.quaver.QuaMapObjMeta import QuaMapObjMode
from reamber.quaver.QuaSvObj import QuaSvObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.base.BpmObj import BpmObj
from reamber.quaver.QuaHitObj import QuaHitObj
from reamber.quaver.QuaHoldObj import QuaHoldObj
from reamber.quaver.QuaBpmObj import QuaBpmObj
from reamber.quaver.lists import QuaBpmList
from reamber.quaver.lists import QuaNotePkg
from reamber.quaver.lists import QuaSvList
from reamber.quaver.lists.notes.QuaHitList import QuaHitList
from reamber.quaver.lists.notes.QuaHoldList import QuaHoldList
from typing import List


class OsuToQua:
    @staticmethod
    def convert(osu: OsuMapObj) -> QuaMapObj:
        """ Converts Osu to a Qua Map

        :param osu: Osu Map
        :return: Quaver map
        """
        assert osu.circleSize == 4 or osu.circleSize == 7

        hits: List[QuaHitObj] = []
        holds: List[QuaHoldObj] = []

        for hit in osu.notes.hits():
            hits.append(QuaHitObj(offset=hit.offset, column=hit.column))
        for hold in osu.notes.holds():
            holds.append(QuaHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

        bpms: List[BpmObj] = []
        svs: List[QuaSvObj] = []

        for bpm in osu.bpms:
            bpms.append(QuaBpmObj(offset=bpm.offset, bpm=bpm.bpm))

        for sv in osu.svs:
            svs.append(QuaSvObj(offset=sv.offset, multiplier=sv.multiplier))

        qua: QuaMapObj = QuaMapObj(
            audioFile=osu.audioFileName,
            title=osu.titleUnicode,
            mode=QuaMapObjMode.str(int(osu.circleSize)),
            artist=osu.artistUnicode,
            creator=osu.creator,
            backgroundFile=osu.backgroundFileName,
            songPreviewTime=osu.previewTime,
            notes=QuaNotePkg(hits=QuaHitList(hits),
                             holds=QuaHoldList(holds)),
            bpms=QuaBpmList(bpms),
            svs=QuaSvList(svs)
        )

        return qua



""" Can't find a better name for this without it being too vague or generic """

from typing import overload

from reamber.osu.OsuMapObj import OsuMapObj, MapObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.quaver.QuaMapObj import QuaMapObj

from copy import deepcopy
from reamber.algorithms.analysis.playField.PlayField import PlayField


def ss(self, pf: PlayField) -> PlayField:
    return PlayField(self)


@overload
def rate(m: OsuMapObj, by: float, inplace: bool = False) -> OsuMapObj: ...
@overload
def rate(m: QuaMapObj, by: float, inplace: bool = False) -> QuaMapObj: ...
@overload
def rate(s: SMMapSetObj, by: float, inplace: bool = False) -> SMMapSetObj: ...
def rate(m: MapObj, by: float, inplace: bool = False) -> MapObj:
    """ Speeds up the map by rate specified.

    :param m: The map
    :param by: The by, 0.8 means 80% speed, 1.2 means 120% speed
    :param inplace: If True, modifies the map in place without returning
    """
    assert 0 < by, "by cannot be negative or zero"

    m_ = m if inplace else deepcopy(m)

    if isinstance(m_, OsuMapObj):
        for bpm in m_.bpms:
            bpm.offset /= by
            bpm.bpm *= by
        for sv in m_.svs:
            sv.offset /= by
        for hit in m_.notes.hits():
            hit.offset /= by
        for hold in m_.notes.holds():
            hold.offset /= by
            hold.length /= by
        for samp in m_.samples:
            samp.offset /= by
        m_.previewTime /= by
        # TODO: Implement rating up of storyboard if implemented

    elif isinstance(m_, QuaMapObj):
        for bpm in m_.bpms:
            bpm.offset /= by
            bpm.bpm /= by
        for sv in m_.svs:
            sv.offset /= by
        for hit in m_.notes.hits():
            hit.offset /= by
        for hold in m_.notes.holds():
            hold.offset /= by
            hold.length /= by
        m_.songPreviewTime /= by

    elif isinstance(m_, SMMapSetObj):
        for map in m_.maps:
            for bpm in map.bpms:
                bpm.offset *= by
                bpm.bpm /= by
        m_.sampleStart /= by
        m_.sampleLength /= by

    return None if inplace else m_



from reamber.osu.lists.OsuSvList import OsuSvList
from reamber.osu.lists.OsuBpmList import OsuBpmList
from reamber.osu.lists.OsuNotePkg import OsuNotePkg

__all__ = ['OsuSvList', 'OsuNotePkg', 'OsuBpmList']



from reamber.osu.OsuMapObj import OsuMapObj
from reamber.sm.SMMapSetObj import SMMapObj
from reamber.o2jam.O2JMapObj import O2JMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from typing import overload, List, Dict


@overload
def scrollSpeed(m: OsuMapObj, centerBpm: float = None) -> List[Dict[str, float]]: ...
@overload
def scrollSpeed(m: SMMapObj, centerBpm: float = None) -> List[Dict[str, float]]: ...
@overload
def scrollSpeed(m: O2JMapObj, centerBpm: float = None) -> List[Dict[str, float]]: ...
@overload
def scrollSpeed(m: QuaMapObj, centerBpm: float = None) -> List[Dict[str, float]]: ...
def scrollSpeed(m: QuaMapObj, centerBpm: float = None) -> List[Dict[str, float]]:
    """ Evaluates the scroll speed based on mapType

    e.g. if BPM == 200.0 and CenterBPM == 100.0, it'll return {'offset': X, 'speed': 2.0}

    :param m: The Map Object
    :param centerBpm: The bpm to zero calculations on. If None, it'll just be the multiplication of bpm and sv.
    :return: Returns a list dict of keys offset and speed
    """

    # This automatically calculates the center BPM
    # Bpm Activity implicitly sorts
    if centerBpm is None: centerBpm = 1

    if isinstance(m, SMMapObj) or isinstance(m, O2JMapObj):  # SM doesn't have SV
        return [dict(offset=bpm.offset, speed=bpm.bpm/centerBpm) for bpm in m.bpms]

    svPairs = [(offset, multiplier) for offset, multiplier in zip(m.svs.sorted().offsets(), m.svs.multipliers())]
    bpmPairs = [(offset, bpm) for offset, bpm in zip(m.bpms.offsets(), m.bpms.bpms())]

    currBpmIter = 0
    nextBpmOffset = None if len(bpmPairs) == 1 else bpmPairs[1][0]
    speedList = []

    for offset, sv in svPairs:
        while offset < bpmPairs[0][0]:  # Offset cannot be less than the first bpm
            continue
        # Guarantee that svOffset is after first bpm
        if nextBpmOffset and offset >= nextBpmOffset:
            currBpmIter += 1
            if currBpmIter != len(bpmPairs):
                nextBpmOffset = bpmPairs[currBpmIter][0]
            else:
                nextBpmOffset = None
        speedList.append(dict(offset=offset, speed=bpmPairs[currBpmIter][1] * sv / centerBpm))

    return speedList



import unittest
from tests.test.RSC_PATHS import *

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj

from reamber.algorithms.mutate.rate import rate
import os

class TestMutateRate(unittest.TestCase):
    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_PLANET_SHAPER)
        rate(m, 2.0, inplace=True)
        m.writeFile("out.osu")

    def test_qua(self):
        m = QuaMapObj()
        m.readFile(QUA_NEURO_CLOUD)
        rate(m, 2.0, inplace=True)
        m.writeFile("out.qua")

    def test_sm(self):
        s = SMMapSetObj()
        s.readFile(SM_CARAVAN)
        rate(s, 2.0, inplace=True)
        s.writeFile("out.sm")


if __name__ == '__main__':
    unittest.main()



from reamber.o2jam.O2JMapSetObj import O2JMapSetObj, O2JMapObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.osu.OsuHitObj import OsuHitObj
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.base.BpmObj import BpmObj
from reamber.osu.lists.OsuBpmList import OsuBpmList
from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.osu.lists.notes.OsuHoldList import OsuHoldList
from reamber.osu.lists.notes.OsuHitList import OsuHitList
from typing import List


class O2JToOsu:
    @staticmethod
    def convert(o2j: O2JMapSetObj) -> List[OsuMapObj]:
        """ Converts a Mapset to multiple Osu maps

        Note that a mapset contains maps, so a list would be expected.
        O2JMap conversion is not possible due to lack of O2JMapset Metadata

        :param o2j: O2Jam Mapset
        :return: List of Osu Maps
        """

        osuMapSet: List[OsuMapObj] = []
        for o2jMap in o2j.maps:
            assert isinstance(o2jMap, O2JMapObj)

            hits: List[OsuHitObj] = []
            holds: List[OsuHoldObj] = []

            # Note Conversion
            for hit in o2jMap.notes.hits():
                hits.append(OsuHitObj(offset=hit.offset, column=hit.column))
            for hold in o2jMap.notes.holds():
                holds.append(OsuHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

            bpms: List[BpmObj] = []

            # Timing Point Conversion
            for bpm in o2jMap.bpms:
                bpms.append(OsuBpmObj(offset=bpm.offset, bpm=bpm.bpm))

            # Extract Metadata
            osuMap = OsuMapObj(
                title=o2j.title,
                artist=o2j.artist,
                creator=o2j.creator,
                version=f"Level {o2j.level[o2j.maps.index(o2jMap)]}",
                bpms=OsuBpmList(bpms),
                circleSize=7,
                notes=OsuNotePkg(hits=OsuHitList(hits),
                                 holds=OsuHoldList(holds))
            )
            osuMapSet.append(osuMap)
        return osuMapSet



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.OsuToSM import OsuToSM
from reamber.osu.OsuMapObj import OsuMapObj
# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestOsuToSM(unittest.TestCase):

    # # @profile
    # def test_osu1(self):
    #     # Complex BPM Points
    #     osu = OsuMapObj()
    #     osu.readFile(OSU_CARAVAN)
    #
    #     sm = OsuToSM.convert(osu)
    #     sm.offset -= 325 + 360
    #     sm.music = "caravan.mp3"
    #     sm.banner = "bn.png"
    #     sm.cdTitle = "cdtitle.png"
    #     sm.writeFile("out.sm", alignBpms=True)
    #
    # # @profile
    # def test_osu2(self):
    #     # Stops
    #     osu = OsuMapObj()
    #     osu.readFile(OSU_ESCAPES)
    #
    #     sm = OsuToSM.convert(osu)
    #     sm.offset -= 635 + 575
    #     sm.writeFile("out.sm", alignBpms=True)

    # @profile
    def test_osu3(self):
        # Complex BPM
        osu = OsuMapObj()
        osu.readFile(OSU_GRAVITY)

        sm = OsuToSM.convert(osu)
        sm.offset -= 15 + 41
        sm.music = "Gravity.mp3"
        sm.writeFile("out.sm", alignBpms=True, BEAT_ERROR_THRESHOLD=5.0, BEAT_CORRECTION_FACTOR=0.001)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.base.lists.notes.HoldList import HoldList
from typing import List


class SMHoldList(List[SMHoldObj], HoldList, SMNoteList):

    def _upcast(self, objList: List = None) -> SMHoldList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMHoldList
        """
        return SMHoldList(objList)

    def data(self) -> List[SMHoldObj]:
        return self



from __future__ import annotations
from reamber.quaver.lists.notes.QuaNoteList import QuaNoteList
from reamber.quaver.QuaHitObj import QuaHitObj
from typing import List


class QuaHitList(List[QuaHitObj], QuaNoteList):

    def _upcast(self, objList: List = None) -> QuaHitList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: QuaHitList
        """
        return QuaHitList(objList)

    def data(self) -> List[QuaHitObj]:
        return self



from __future__ import annotations
from reamber.base.HoldObj import HoldObj
from reamber.osu.OsuNoteObjMeta import OsuNoteObjMeta
from dataclasses import dataclass


@dataclass
class OsuHoldObj(HoldObj, OsuNoteObjMeta):
    @staticmethod
    def readString(s: str, keys: int) -> OsuHoldObj or None:
        """ Reads a single line under the [HitObject] Label. This must explicitly be a Hold Object.

        keys must be specified for conversion of code value to actual column."""
        if s.isspace():
            return None

        sComma = s.split(",")
        if len(sComma) < 5:
            return None

        sColon = sComma[-1].split(":")
        if len(sColon) < 6:
            return None

        this = OsuHoldObj()
        this.column = this.xAxisToColumn(int(sComma[0]), keys)
        this.offset = float(sComma[2])
        this.hitsoundSet = int(sComma[4])
        this.length = float(sColon[0]) - this.offset
        this.sampleSet = int(sColon[1])
        this.additionSet = int(sColon[2])
        this.customSet = int(sColon[3])
        this.volume = int(sColon[4])
        this.hitsoundFile = sColon[5]

        return this

    def writeString(self, keys: int) -> str:
        """ Exports a .osu writable string """
        return f"{OsuNoteObjMeta.columnToXAxis(self.column, keys=keys)},{192}," \
               f"{int(self.offset)},{128},{self.hitsoundSet},{int(self.offset + self.length)}:" \
               f"{self.sampleSet}:{self.additionSet}:{self.customSet}:{self.volume}:{self.hitsoundFile}"



from __future__ import annotations
from dataclasses import dataclass, field

from reamber.base.MapObj import MapObj
from reamber.base.RAConst import RAConst
from reamber.base.lists import TimedList
from reamber.o2jam.O2JEventPackage import O2JEventPackage
from reamber.o2jam.lists.O2JNotePkg import O2JNotePkg
from reamber.o2jam.lists.O2JBpmList import O2JBpmList, O2JBpmObj
from reamber.o2jam.lists.notes.O2JHitList import O2JHitList
from reamber.o2jam.lists.notes.O2JHoldList import O2JHoldList
from reamber.o2jam.O2JHitObj import O2JHitObj
from reamber.o2jam.O2JHoldObj import O2JHoldObj
from typing import List, Dict
import logging

log = logging.getLogger(__name__)

@dataclass
class O2JMapObj(MapObj):
    """ This holds a single level of a .ojn file out of a total of three.

    This class only holds the data of notes and bpms. The rest can be found in the parent O2JMapSetObj instance.

    We won't support OJM, see why in O2JMapSetObj. """

    notes: O2JNotePkg = field(default_factory=lambda: O2JNotePkg())
    bpms:  O2JBpmList = field(default_factory=lambda: O2JBpmList())

    def data(self) -> Dict[str, TimedList]:
        """ Gets the notes and bpms as a dictionary """
        return {'notes': self.notes,
                'bpms': self.bpms}

    # noinspection PyUnresolvedReferences
    @staticmethod
    def readPkgs(pkgs: List[O2JEventPackage], initBpm: float) -> O2JMapObj:
        """ Reads a level/map package and returns a O2JMapObj

        :param pkgs: A map package
        :param initBpm: The initial bpm for the map, it's the one located in the meta
        :return:
        """

        """ The main idea here is to get unique measures from notes and find out the real offsets via bpm
        The catch is that we drop them into the eventsNoteDict as a dictionary like
        { 10: 300, 12: 5000 }

        Then we loop through the notes to obtain the correct offsets.

        We do this because we don't want to separate and pair the LNs again since we have to go through this
        algorithm by a sorted offset.
        """
        events = [event for pkg in pkgs for event in pkg.events]
        events.sort(key=lambda x: x.measure)
        notes = [event for event in events if not isinstance(event, O2JBpmObj)]
        noteMeasures = {event.measure for event in notes} | \
                       {event.tailMeasure for event in notes if isinstance(event, O2JHoldObj)}
        noteMeasures = sorted(noteMeasures)
        noteMeasureDict = {}
        bpms = [event for event in events if isinstance(event, O2JBpmObj)]

        currOffset = 0
        currMeasure = 0
        currBpmI = -1
        currBpm = initBpm
        nextBpmMeasure = bpms[0].measure if len(bpms) > 0 else None
        for noteMeasure in noteMeasures:
            if nextBpmMeasure:
                while noteMeasure > nextBpmMeasure:
                    # Move BPM Index up by 1
                    currBpmI += 1

                    # Update offset
                    currOffset += RAConst.minToMSec((bpms[currBpmI].measure - currMeasure) * 4 / currBpm)
                    bpms[currBpmI].offset = currOffset
                    currMeasure = bpms[currBpmI].measure
                    currBpm = bpms[currBpmI].bpm

                    # Check if next one is available
                    if currBpmI + 1 == len(bpms):
                        nextBpmMeasure = None
                        break
                    else:
                        nextBpmMeasure = bpms[currBpmI + 1].measure

            # We add it into the measure: offset dictionary.
            noteMeasureDict[noteMeasure] = \
                currOffset + RAConst.minToMSec(4 * (noteMeasure - currMeasure) / currBpm)

        # We then assign all the offsets here
        for note in notes:
            note.offset = noteMeasureDict[note.measure]
            if isinstance(note, O2JHoldObj):  # Special case for LN.
                note.length = noteMeasureDict[note.tailMeasure] - note.offset

        # We add the missing first BPM here
        bpms.insert(0, O2JBpmObj(offset=0, bpm=initBpm))
        return O2JMapObj(notes=O2JNotePkg(hits=O2JHitList([n for n in notes if isinstance(n, O2JHitObj)]),
                                          holds=O2JHoldList([n for n in notes if isinstance(n, O2JHoldObj)])),
                         bpms=O2JBpmList(bpms))



""" This algorithm finds the offsets for every 1/1 (4th) beat in the maps' bpm list

To support image/image algorithm
"""

from reamber.base.lists.BpmList import BpmList
from reamber.base.RAConst import RAConst
from reamber.osu.lists.OsuBpmList import OsuBpmList
from typing import overload, List


def bpmBeatOffsets(bpms: OsuBpmList, nths: float = 1.0,
                   lastOffset: float = None):  # to convert into BPMList later
    """ Gets all of the nth snap offsets

    For example::

        SEC     1   2   3   4   5   6   7   8   9   10  11  12  13  14  ...
        BPM     15                          7.5                         ...
        SNAP    4/4 1/4 2/4 3/4 4/4 1/4 2/4 4/4 1/8 2/8 3/8 4/8 5/8 6/8 ...
        BEAT    1               2           3
        nths=1  ^               ^           ^
        nths=2  ^       ^       ^       ^   ^               ^
        nths=4  ^   ^   ^   ^   ^   ^   ^   ^       ^       ^       ^
        nths=8  ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^   ^   ^   ^   ^   ^   ^

    * Note: 15 BPM = 1 Beat per 4 seconds, 7.5 = 1 Beat per 8 seconds

    The `^` indicates what offsets will be returned.

    :param bpms: The map's bpm
    :param nths: Specifies the beat's snap, 1 = "1st"s, 4 = "4th"s, 16 = "16th"s
    :param lastOffset: The last offset to consider, if None, it uses the last BPM
    """
    offsets: List[float] = []
    bpms_ = bpms.sorted().data()

    currBpmI   = 0
    currOffset = bpms_[currBpmI].offset
    currBpm    = bpms_[currBpmI]
    snapLength = RAConst.minToMSec(1 / nths / currBpm.bpm)
    if lastOffset is None: lastOffset = bpms_[-1].offset

    nextBpm = None if currBpmI + 1 == len(bpms_) else bpms_[currBpmI + 1]

    while currOffset <= lastOffset:
        offsets.append(currOffset)
        currOffset += snapLength
        if nextBpm and currOffset > nextBpm.offset:
            currBpm = nextBpm
            currOffset = nextBpm.offset
            snapLength = RAConst.minToMSec(1 / nths / currBpm.bpm)
            currBpmI += 1
            nextBpm = None if currBpmI + 1 == len(bpms_) else bpms_[currBpmI + 1]
            continue

    return offsets



from __future__ import annotations
from typing import List
from reamber.base.lists.BpmList import BpmList
from reamber.sm.SMBpmObj import SMBpmObj


class SMBpmList(BpmList):

    def _upcast(self, objList: List = None) -> SMBpmList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMBpmList
        """
        return SMBpmList(objList)

    def data(self) -> List[SMBpmObj]:
        return self



from __future__ import annotations
from reamber.base.lists.TimedList import TimedList
from reamber.osu.OsuSvObj import OsuSvObj
from typing import List


class OsuSvList(List[OsuSvObj], TimedList):

    def _upcast(self, objList: List = None) -> OsuSvList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: OsuSvList
        """
        return OsuSvList(objList)

    def data(self) -> List[OsuSvObj]:
        return self

    def multipliers(self) -> List[float]:
        """ Gets all Scroll Velocity values """
        return self.attribute('multiplier')



from reamber.sm.lists.notes.SMRollList import SMRollList
from reamber.sm.lists.notes.SMHoldList import SMHoldList
from reamber.sm.lists.notes.SMFakeList import SMFakeList
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.lists.notes.SMKeySoundList import SMKeySoundList
from reamber.sm.lists.notes.SMHitList import SMHitList
from reamber.sm.lists.notes.SMLiftList import SMLiftList
from reamber.sm.lists.notes.SMMineList import SMMineList


__all__ = ['SMRollList', 'SMHoldList', 'SMFakeList', 'SMNoteList', 'SMKeySoundList',
           'SMHitList', 'SMLiftList', 'SMMineList']



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestCombine(unittest.TestCase):

    def test(self):

        t = SvPkg([SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)]),
                   SvSequence([SvObj(100, 3.0), SvObj(200, 4.0)]),
                   SvSequence([SvObj(150, 5.0), SvObj(300, 6.0)])])\
            .combine(combineMethod=SvPkg.CombineMethod.DROP_BY_POINT, combineMethodWindow=1)

        self.assertEqual(len(t), 5)

        t = SvPkg([SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)]),
                   SvSequence([SvObj(100, 3.0), SvObj(200, 4.0)]),
                   SvSequence([SvObj(150, 5.0), SvObj(300, 6.0)])])\
            .combine(combineMethod=SvPkg.CombineMethod.DROP_BY_BOUND, combineMethodWindow=1)

        self.assertEqual(len(t), 4)

        t = SvPkg([SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)]),
                   SvSequence([SvObj(100, 3.0), SvObj(200, 4.0)]),
                   SvSequence([SvObj(150, 5.0), SvObj(300, 6.0)])])\
            .combine(combineMethod=SvPkg.CombineMethod.IGNORE, combineMethodWindow=1)

        self.assertEqual(len(t), 6)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from reamber.base.lists.NotePkg import NotePkg
from reamber.sm.lists.notes import *
from typing import Dict, overload


class SMNotePkg(NotePkg):
    """ This package holds hits, holds, rolls, mines, lifts, fakes, keySounds for each SMMapObj """

    @overload
    def __init__(self): ...
    @overload
    def __init__(self, dataDict: Dict[str, SMNoteList]): ...
    @overload
    def __init__(self, hits: SMHitList, holds: SMHoldList, rolls: SMRollList, mines: SMMineList,
                 lifts: SMLiftList, fakes: SMFakeList, keySounds: SMKeySoundList): ...
    def __init__(self, dataDict=None, hits=None, holds=None, rolls=None, mines=None, lifts=None, fakes=None,
                 keySounds=None):
        """ Initialize a package,

        Can initialize with either overloaded method.

        :param dataDict: The data dictionary, it'll be directly assigned to dataDict. The names must explicitly match
        :param hits: The hits as a SMHitList
        :param holds: The holds as a SMHoldList
        :param rolls: The holds as a SMRollList
        :param mines: The holds as a SMMineList
        :param lifts: The holds as a SMLiftList
        :param fakes: The holds as a SMFakeList
        :param keySounds: The holds as a SMKeySoundList
        """
        if dataDict is not None: self.dataDict = dataDict
        elif hits is not None: self.dataDict = {'hits': hits, 'holds': holds, 'rolls': rolls, 'mines': mines,
                                                'lifts': lifts, 'fakes': fakes, 'keySounds': keySounds}
        else: self.dataDict: Dict[str, SMNoteList] = {'hits': SMHitList(), 'holds': SMHoldList(), 'rolls': SMRollList(),
                                                      'mines': SMMineList(), 'lifts': SMLiftList(),
                                                      'fakes': SMFakeList(), 'keySounds': SMKeySoundList()}

    def _upcast(self, dataDict: Dict[str, SMNoteList]) -> SMNotePkg:
        """ This is to facilitate inherited functions to work """
        return SMNotePkg(dataDict=dataDict)

    def __iter__(self):
        """ Yields the Dictionary item by item """
        yield from self.dataDict

    def data(self) -> Dict[str, SMNoteList]:
        """ Returns the data dictionary of lists """
        return self.dataDict

    # noinspection PyTypeChecker
    def hits(self) -> SMHitList:
        """ Returns the hitList from the dictionary """
        return self.dataDict['hits']
    # noinspection PyTypeChecker
    def holds(self) -> SMHoldList:
        """ Returns the holdList from the dictionary """
        return self.dataDict['holds']
    # noinspection PyTypeChecker
    def rolls(self) -> SMRollList:
        """ Returns the rollList from the dictionary """
        return self.dataDict['rolls']
    # noinspection PyTypeChecker
    def mines(self) -> SMMineList:
        """ Returns the mineList from the dictionary """
        return self.dataDict['mines']
    # noinspection PyTypeChecker
    def lifts(self) -> SMLiftList:
        """ Returns the liftList from the dictionary """
        return self.dataDict['lifts']
    # noinspection PyTypeChecker
    def fakes(self) -> SMFakeList:
        """ Returns the fakeList from the dictionary """
        return self.dataDict['fakes']
    # noinspection PyTypeChecker
    def keySounds(self) -> SMKeySoundList:
        """ Returns the keySoundList from the dictionary """
        return self.dataDict['keySounds']






from reamber.base.lists.notes.NoteList import NoteList
from typing import List, Type
from abc import ABC

from reamber.o2jam.O2JNoteObjMeta import O2JNoteObjMeta


class O2JNoteList(NoteList, ABC):
    def data(self) -> List[Type[O2JNoteObjMeta]]: pass



"""

All classes/files in this folder are generators.

That means their purpose is to ease the creation of common SV Sequences.

"""


from reamber.algorithms.generate.sv.generators.svNormalizeBpm import svNormalizeBpm
from reamber.algorithms.generate.sv.generators.svFuncSequencer import svFuncSequencer
from reamber.algorithms.generate.sv.generators.svOsuMeasureLineA import svOsuMeasureLineA
from reamber.algorithms.generate.sv.generators.svOsuMeasureLineB import svOsuMeasureLineB


__all__ = ['svNormalizeBpm', 'svFuncSequencer', 'svOsuMeasureLineA', 'svOsuMeasureLineB']


from reamber.o2jam.O2JMapSetObj import O2JMapSetObj, O2JMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.quaver.QuaHitObj import QuaHitObj
from reamber.quaver.QuaHoldObj import QuaHoldObj
from reamber.quaver.QuaBpmObj import QuaBpmObj
from reamber.base.BpmObj import BpmObj
from reamber.quaver.lists.QuaNotePkg import QuaNotePkg
from reamber.quaver.lists.QuaBpmList import QuaBpmList
from reamber.quaver.lists.notes.QuaHoldList import QuaHoldList
from reamber.quaver.lists.notes.QuaHitList import QuaHitList
from typing import List


class O2JToQua:
    @staticmethod
    def convert(o2j: O2JMapSetObj) -> List[QuaMapObj]:
        """ Converts a Mapset to multiple Quaver maps

        Note that a mapset contains maps, so a list would be expected.
        O2JMap conversion is not possible due to lack of O2JMapset Metadata

        :param o2j: O2Jam Mapset
        :return: List of Quaver Maps
        """
        quaMapSet: List[QuaMapObj] = []
        for o2jMap in o2j.maps:
            assert isinstance(o2jMap, O2JMapObj)
            hits: List[QuaHitObj] = []
            holds: List[QuaHoldObj] = []

            # Note Conversion
            for hit in o2jMap.notes.hits():
                hits.append(QuaHitObj(offset=hit.offset, column=hit.column))
            for hold in o2jMap.notes.holds():
                holds.append(QuaHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

            bpms: List[BpmObj] = []

            # Timing Point Conversion
            for bpm in o2jMap.bpms:
                bpms.append(QuaBpmObj(offset=bpm.offset, bpm=bpm.bpm))

            # Extract Metadata
            quaMap = QuaMapObj(
                title=o2j.title,
                artist=o2j.artist,
                creator=o2j.creator,
                difficultyName=f"Level {o2j.level[o2j.maps.index(o2jMap)]}",
                bpms=QuaBpmList(bpms),
                notes=QuaNotePkg(hits=QuaHitList(hits),
                                 holds=QuaHoldList(holds))
            )
            quaMapSet.append(quaMap)
        return quaMapSet



""" The purpose of having separate classes for these objects is to facilitate functions that deal with the whole list

Note: If it's possible to classify a certain object under notes or

E.g.
Instead of having to do a for loop to shift offset ...
    for obj in map:
        obj.offset += 100

We can just call a function
    map.notes.addOffset(100)

There are also other helper classifiers such as MapObjDataFrame, which indicates that the class can be coerced into
a pandas DataFrame.

Convention
When inheriting from this, place everything in lists package.
Names inheriting should be plural, to be clear
Also, inherit from the List[Obj] with the appropriate __init__

E.g. class BpmList(List[BpmObj], TimedList, ...)
        def __init__(self, *args):
            list.__init__(self, *args)

Unlike most things in the repository, this is not a @dataclass

"""

from reamber.base.lists.BpmList import BpmList
from reamber.base.lists.TimedList import TimedList
from reamber.base.lists.NotePkg import NotePkg

__all__ = ['BpmList', 'TimedList', 'NotePkg']

""" Instructions on subclassing

Whenever you want to create a class that inherits from this, there are a few things to take note of

1.1 Inherit from List[Obj]
1.2 You need to declare the def __init__ for initialization of List
    This is the generic way to do it
        def __init__(self, *args):
            list.__init__(self, *args)

2.1 Union of List
    Consider this
    svs: Union[OsuSvList, List[OsuSvObj]] = field(default_factory=lambda: OsuSvList())
    
    Union here is required else PyCharm will complain that you cannot allocate a List[OsuSvObj] to svs
    
    Of course it's not valid, that's where _recast comes into play
    
2.2 _recast
    In the base class, __post_init__ will call _recast after initializing.
    You just need to override _recast.
    I made it a separate function because there may be multiple super().__post_init__ s
    
    def _recast(self) -> None:
        super()._recast()  # Remember to call the super
        self.svs = OsuSvList(self.svs)

"""




from __future__ import annotations
from dataclasses import asdict
from abc import abstractmethod, ABC
from typing import List, Tuple, Type
import pandas as pd
from copy import deepcopy

""" Criterion
The derived object must be:
1. A List of @dataclass
2. DataFrame-able (implied in 1.) <See df(self) on how it implicitly defines a dataclass DF
"""

""" Convention
The idea of most functions here is to be able to chain continuously, then get the result using data() or offset()
obj.func().funcOther().data()

The class must also be able to be casted into a DataFrame
"""


class TimedList(ABC):
    """ A class to handle all derives' offset-related functions.

    All derived class must inherit from a list of their singular type
    """

    def __init__(self, *args):
        if args: list.__init__(*args)
        else: list.__init__([])

    @abstractmethod
    def data(self) -> List:
        """ The method to grab the data from derived classes """
        pass

    @abstractmethod
    def _upcast(self, objList: List = None):
        """ The method to upcast to the derived class

        The premise of upcast is that if I casted all functions to this current class, it'll end up using the absmethod
        data(self), which will return None.

        Hence the derived classes should also implement upcast
        """
        pass

    def df(self) -> pd.DataFrame:
        """ The object itself must be df-able """
        return pd.DataFrame([asdict(obj) for obj in self.data()])

    def deepcopy(self) -> TimedList:
        return deepcopy(self)

    def sorted(self, reverse: bool = False, inplace: bool = False) -> TimedList:
        """ Sorts the list by offset

        :param reverse: Whether to sort in reverse or not
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.__init__(sorted(self.data(), key=lambda tp: tp.offset, reverse=reverse))
        else: return self._upcast(sorted(self.data(), key=lambda tp: tp.offset, reverse=reverse))

    def between(self, lowerBound, upperBound, includeEnds=True, inplace: bool = False) -> TimedList:
        """ Trims the list between specified bounds

        :param lowerBound: The lower bound in milliseconds
        :param upperBound: The upper bound in milliseconds
        :param includeEnds: Whether to include the bound ends. \
            Use after and before if you need to only include 1 end
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.before(lowerBound, includeEnds, inplace=False)\
                        .after(upperBound, includeEnds, inplace=False)
        else: return self.before(lowerBound, includeEnds, inplace=False)\
                         .after(upperBound, includeEnds, inplace=False)

    def after(self, offset: float, includeEnd : bool = False, inplace: bool = False) -> TimedList:
        """ Trims the list after specified offset

        :param offset: The lower bound in milliseconds
        :param includeEnd: Whether to include the end
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.__init__([obj for obj in self.data() if obj.offset >= offset]) if includeEnd else \
                    self.__init__([obj for obj in self.data() if obj.offset > offset])
        else: return self._upcast([obj for obj in self.data() if obj.offset >= offset]) if includeEnd else \
                     self._upcast([obj for obj in self.data() if obj.offset > offset])

    def before(self, offset: float, includeEnd : bool = False, inplace: bool = False) -> TimedList:
        """ Trims the list before specified offset

        :param offset: The upper bound in milliseconds
        :param includeEnd: Whether to include the end
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.__init__([obj for obj in self.data() if obj.offset <= offset]) if includeEnd else \
                    self.__init__([obj for obj in self.data() if obj.offset < offset])
        else: return self._upcast([obj for obj in self.data() if obj.offset <= offset]) if includeEnd else \
                     self._upcast([obj for obj in self.data() if obj.offset < offset])

    def attribute(self, method: str) -> List:
        """ Calls each obj's method with eval. Specify method with a string.

        :param method: The method to call, the string must be **EXACT**
        :return: Returns a List of the result
        """
        expression = f"_.{method}"
        asFunc = eval('lambda _: ' + expression)

        return [asFunc(_) for _ in self.data()]
        # The above is faster for some reason
        # return [eval(f"_.{method}") for _ in self.data()]

    def instances(self, instanceOf: Type, inplace: bool = False) -> TimedList:
        """ Gets all instances that match the instanceOf type

        :param instanceOf: The type to match
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.__init__([obj for obj in self.data() if isinstance(obj, instanceOf)])
        else: return self._upcast([obj for obj in self.data() if isinstance(obj, instanceOf)])

    def offsets(self) -> List[float]:
        """ Gets all offsets of the objects """
        return [obj.offset for obj in self.data()]

    def setOffsets(self, offsets: List[float]):
        """ Sets all offsets with a List """
        for offset, obj in zip(offsets, self.data()):
            obj.offset = offset

    def addOffset(self, by: float, inplace: bool = False) -> TimedList:
        """ Adds offset to all object

        :param by: The offset to move by
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: d = self.data()
        else: d = self.data()[:]
        for i, obj in enumerate(d):
            obj.offset += by
            d[i] = obj
        if not inplace: return self._upcast(d)

    def multOffset(self, by: float, inplace: bool = False) -> TimedList:
        """ Adds offset to all object

        :param by: The offset to move by
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: d = self.data()
        else: d = self.data()[:]
        for i, obj in enumerate(d):
            obj.offset *= by
            d[i] = obj
        if not inplace: return self._upcast(d)

    def lastOffset(self) -> float:
        """ Get Last Note Offset """
        if len(self.data()) == 0: return 0.0
        return max([obj.offset for obj in self.data()])

    def firstOffset(self) -> float:
        """ Get First Note Offset """
        if len(self.data()) == 0: return float("inf")
        return min([obj.offset for obj in self.data()])

    def firstLastOffset(self) -> Tuple[float, float]:
        """ Get First and Last Note Offset
        This is slightly faster than separately calling the singular functions since it sorts once only
        """
        if len(self.data()) == 0: return 0.0, float('inf')
        obj = self.sorted().data()
        return obj[0].offset, obj[-1].offset

    def duration(self) -> float:
        """ Gets the total duration of this list """
        first, last = self.firstLastOffset()
        return last - first

    def moveStartTo(self, to: float, inplace:bool = False) -> TimedList:
        """ Moves the start of this list to a specific offset

        :param to: The offset to move it to
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        first = self.firstOffset()
        return self.addOffset(to - first, inplace=inplace)

    def moveEndTo(self, to: float, inplace:bool = False) -> TimedList:
        """ Moves the end of this list to a specific offset

        :param to: The offset to move it to
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        last = self.lastOffset()
        return self.addOffset(to - last, inplace=inplace)

#
# def generateAbc(singularType: Type = None, data=True, upcast=True):
#     """ This factory creates a decorator that sets the basic necessities for anything deriving from a mapobjBase
#     It adds __init__, data, and _upcast
#     :param singularType: This must be declared if data is true
#     :param data: Default True, generates the data
#     :param upcast: Default True, generates the upcast
#     :return: """
#
#     def wrapper(cls):
#         def _data(self) -> List[singularType]:
#             return self
#
#         def _upcast(self, m: List = None) -> cls:
#             if m is None: m = []
#             return cls(m)
#
#         if data:   setattr(cls, 'data', _data)
#         if upcast: setattr(cls, '_upcast', _upcast)
#
#         return cls
#     return wrapper



from __future__ import annotations
from typing import List
from reamber.base.lists.BpmList import BpmList
from reamber.o2jam.O2JBpmObj import O2JBpmObj


class O2JBpmList(BpmList):

    def _upcast(self, objList: List = None) -> O2JBpmList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: O2JBpmList
        """
        return O2JBpmList(objList)

    def data(self) -> List[O2JBpmObj]:
        return self



import matplotlib.pyplot as plt
import matplotlib.axes as axes
from typing import overload

import pandas as pd

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.sm.SMMapSetObj import SMMapObj
from reamber.o2jam.O2JMapObj import O2JMapObj
from reamber.quaver.QuaMapObj import QuaMapObj

@overload
def nps(m: O2JMapObj, binSize: int = 1000) -> pd.DataFrame: ...
@overload
def nps(m: OsuMapObj, binSize: int = 1000) -> pd.DataFrame: ...
@overload
def nps(m: QuaMapObj, binSize: int = 1000) -> pd.DataFrame: ...
@overload
def nps(m: SMMapObj, binSize: int = 1000) -> pd.DataFrame: ...
def nps(m: SMMapObj, binSize: int = 1000) -> pd.DataFrame:
    """ Gets the NPS as a DataFrame

    :param m: The MapObj or any variant
    :param binSize: The size of the binning
    """

    dfMaster = None

    for k, l in m.notes.data().items():
        if len(l.data()) == 0: continue
        # Fence post issue, last offset will be cut short, so we add a bin to cover the end
        dfCut = pd.cut(l.df()['offset'], bins=list(range(0, int(l.lastOffset()) + binSize, binSize)))
        dfCut = dfCut.groupby(dfCut).count()
        df = pd.DataFrame({f"{k}": dfCut.values / (binSize / 1000)})
        df = df.reset_index(inplace=False).rename(columns={'index': 'offset'}, inplace=False)
        df['offset'] *= binSize
        if dfMaster is None: dfMaster = df
        else: dfMaster = dfMaster.merge(df)

    return dfMaster

@overload
def npsPlot(m: O2JMapObj, widthPx=1000, heightPx=200, dpi=100, binSize=1000) -> None: ...
@overload
def npsPlot(m: OsuMapObj, widthPx=1000, heightPx=200, dpi=100, binSize=1000) -> None: ...
@overload
def npsPlot(m: QuaMapObj, widthPx=1000, heightPx=200, dpi=100, binSize=1000) -> None: ...
@overload
def npsPlot(m: SMMapObj, widthPx=1000, heightPx=200, dpi=100, binSize=1000) -> None: ...
def npsPlot(m: SMMapObj, widthPx=1000, heightPx=200, dpi=100, binSize=1000) -> None:
    """ This creates an NPS plot. The peaks and troughs may differ depending on binSize

    :param dpi: dpi
    :param heightPx: Height in pixels
    :param widthPx: Width in pixels
    :param m: The MapObj or any variant
    :param binSize: The size of the binning
    """
    df = nps(m, binSize=binSize)
    df.set_index(df['offset'], inplace=True)
    df.drop('offset', axis='columns', inplace=True)
    subplot = df.plot(kind='bar', stacked=True, width=1.0)
    ax: axes.Axes = subplot.axes
    ax.set_xticks([])
    ax.set_ylabel("NPS")
    ax.set_xlabel("")
    fig = ax.get_figure()
    fig.set_size_inches(widthPx / dpi, heightPx / dpi)
    plt.tight_layout()




from reamber.base.BpmObj import BpmObj
from reamber.base.lists.TimedList import TimedList
from reamber.base.TimedObj import TimedObj
from typing import Tuple, List

def activity(tl: TimedList, lastOffset:float = None) -> List[Tuple[TimedObj, float]]:
    """ Calculates how long each Timed Object is active. Implicitly sorts object by offset

    For example:

    The algorithm calculates this::

        SEC 1   2   3   4   5   6   7   8   9
        BPM 100 ------> 200 --> 300 -------->

    returns [(TimedObj<100>, 3000), (TimedObj<200>, 2000), (TimedObj<300>, 3000)]

    :param tl: Timed List
    :param lastOffset: Last offset, if None, uses TimedObj.lastOffset()
    :return A List of Tuples in the format [(TimedObj, Activity In ms), ...]
    """

    if lastOffset is None: lastOffset = tl.lastOffset()

    # Describes the BPM and Length of it active
    # e.g. [(120.0, 2000<ms>), (180.0, 1000<ms>), ...]
    acts: List[Tuple[BpmObj, float]] = []

    for obj in tl.sorted(reverse=True).data():
        if obj.offset >= lastOffset:
            acts.append((obj, 0.0))  # If the BPM doesn't cover any notes it is inactive
        else:
            acts.append((obj, lastOffset - obj.offset))
            lastOffset = obj.offset
    return list(reversed(acts))




class SMConst:
    HIT_STRING: str = "1"
    HOLD_STRING_HEAD: str = "2"
    HOLD_STRING_TAIL: str = "3"
    LIFT_STRING: str = "L"
    KEYSOUND_STRING: str = "K"
    FAKE_STRING: str = "F"
    MINE_STRING: str = "M"
    ROLL_STRING_HEAD: str = "4"
    ROLL_STRING_TAIL: str = "3"



from __future__ import annotations
from reamber.base.lists.notes.NoteList import NoteList
from typing import List, Type
from abc import ABC

from reamber.quaver.QuaNoteObjMeta import QuaNoteObjMeta


class QuaNoteList(NoteList, ABC):

    def _upcast(self, objList: List = None) -> QuaNoteList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: QuaNoteList
        """
        return QuaNoteList(objList)

    def data(self) -> List[Type[QuaNoteObjMeta]]: pass

    def keySoundsList(self):
        return self.attribute('keySounds')



class OsuSampleSet:
    """ Osu SampleSet are a "enum" of available hitsounds.

    Not to be confused with OsuSampleObj, where that's a class for samples under [Events]
    """
    AUTO: int = 0
    NORMAL: int = 1
    SOFT: int = 2
    DRUM: int = 3

    @staticmethod
    def toString(sampleSet: int) -> str:
        """ Converts a integer to a string representing SampleSet """
        if sampleSet == OsuSampleSet.AUTO:
            return "None"
        elif sampleSet == OsuSampleSet.NORMAL:
            return "Normal"
        elif sampleSet == OsuSampleSet.SOFT:
            return "Soft"
        elif sampleSet == OsuSampleSet.DRUM:
            return "Drum"
        else:
            return "Invalid"

    @staticmethod
    def fromString(sampleSet: str) -> int:
        """ Converts a string to an integer """
        if sampleSet == "None":
            return OsuSampleSet.AUTO
        elif sampleSet == "Normal":
            return OsuSampleSet.NORMAL
        elif sampleSet == "Soft":
            return OsuSampleSet.SOFT
        elif sampleSet == "Drum":
            return OsuSampleSet.DRUM
        else:
            return -1



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMRollObj import SMRollObj
from typing import List


class SMRollList(List[SMRollObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMRollList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMRollList
        """
        return SMRollList(objList)

    def data(self) -> List[SMRollObj]:
        return self

    # Copied from Holds, don't think it's worth splitting this further to just reduce repeated code
    def lengths(self) -> List[float]:
        return self.attribute('length')

    def tailOffsets(self) -> List[float]:
        return [obj() for obj in self.attribute('tailOffset')]



from __future__ import annotations
from typing import List
from reamber.osu.OsuSampleObj import OsuSampleObj
from reamber.base.lists.TimedList import TimedList


class OsuSampleList(List[OsuSampleObj], TimedList):

    def _upcast(self, objList: List = None) -> OsuSampleList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: OsuSampleList
        """
        return OsuSampleList(objList)

    def data(self) -> List[OsuSampleObj]:
        return self

    def sampleFiles(self) -> List[str]:
        return self.attribute('sampleFile')

    def volumes(self) -> List[int]:
        return self.attribute('volume')



from reamber.base.NoteObj import NoteObj
from dataclasses import dataclass


@dataclass
class HoldObj(NoteObj):
    """ A holdable timed object with a specified length.

    We only store the length, the tail offset is calculated. """
    length: float = 0.0

    def tailOffset(self) -> float:
        """ Gets the offset for the tail """
        return self.offset + self.length



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestFit(unittest.TestCase):

    def testFit(self):
        # Test Fitting
        seq = SvSequence([SvObj(0, 0.5), SvObj(50, 1.5), SvObj(100, 1.0)])
        seq = SvPkg.fit(seq, [0, 100, 200, 400, 600]).combine(SvPkg.CombineMethod.DROP_BY_POINT)

        self.assertEqual(len(seq), 9)


if __name__ == '__main__':
    unittest.main()



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestCross(unittest.TestCase):

    def test(self):
        # Test Cross
        seq = SvSequence([SvObj(0, 0.5), SvObj(100, 2.0), SvObj(200, 3.0), SvObj(400, 2.0)])
        seq2 = SvSequence([SvObj(50, 0.1), SvObj(100, 10.0), SvObj(250, 5.0), SvObj(500, 0.3)])
        seq.crossWith(seq2, inplace=True)

        self.assertAlmostEqual(seq[0].multiplier, 0.5)
        self.assertAlmostEqual(seq[1].multiplier, 20.0)
        self.assertAlmostEqual(seq[2].multiplier, 30.0)
        self.assertAlmostEqual(seq[3].multiplier, 10.0)

    def test2(self):
        # Test Mutual Cross
        seq1 = SvSequence([SvObj(0, 0.5), SvObj(100, 2.0), SvObj(200, 3.0), SvObj(400, 2.0)])
        seq2 = SvSequence([SvObj(50, 0.1), SvObj(100, 10.0), SvObj(250, 5.0), SvObj(500, 0.2)])
        seq = SvPkg.crossMutualWith(seq1, seq2).combine(SvPkg.CombineMethod.DROP_BY_POINT)

        self.assertAlmostEqual(seq[0].multiplier, 0.5)
        self.assertAlmostEqual(seq[1].multiplier, 0.05)
        self.assertAlmostEqual(seq[2].multiplier, 20.0)
        self.assertAlmostEqual(seq[3].multiplier, 30.0)
        self.assertAlmostEqual(seq[4].multiplier, 15.0)
        self.assertAlmostEqual(seq[5].multiplier, 10.0)
        self.assertAlmostEqual(seq[6].multiplier, 0.4)




if __name__ == '__main__':
    unittest.main()



from reamber.base.BpmObj import BpmObj
from reamber.base.MapObj import MapObj
from reamber.algorithms.analysis.generic.activity import activity
from typing import Tuple, List


def bpmActivity(m: MapObj) -> List[Tuple[BpmObj, float]]:
    """ Calculates how long the Bpm is active. Implicitly sorts BPM

    For example

    The algorithm calculates this::

        SEC 1   2   3   4   5   6   7   8   9
        BPM 100 ------> 200 --> 300 -------->

    returns [(BPMPoint<100>, 3000), (BPMPoint<200>, 2000), (BPMPoint<300>, 3000)]

    :param m: Map Object
    :return: A List of Tuples in the format [(BPMPoint, Activity In ms), ...]
    """

    # Guaranteed to return BpmObj.
    # noinspection PyTypeChecker
    return activity(m.bpms, m.notes.lastOffset())



# from reamber.algorithms.analysis.note.density import *
# from reamber.algorithms.analysis.note.nps import *
# from reamber.algorithms.analysis.generic import *
# from reamber.algorithms.analysis.describe.describe import *
# from reamber.algorithms.analysis.describe.meta import *
# # from reamber.algorithms.analysis._asDataFrame import asDataFrame
# from reamber.algorithms.analysis.bpm import aveBpm, bpmActivity
#
# import reamber.algorithms.analysis.bpm
# import reamber.algorithms.analysis.describe
# import reamber.algorithms.analysis.note
# import reamber.algorithms.analysis.bpm
#
# __all__ = ['bpm', 'density', 'asDataFrame', 'describePrint', 'describePlot', 'bpmActivity', 'aveBpm']



from __future__ import annotations
from reamber.quaver.lists.notes.QuaNoteList import QuaNoteList
from reamber.quaver.QuaHoldObj import QuaHoldObj
from reamber.base.lists.notes.HoldList import HoldList
from typing import List


class QuaHoldList(List[QuaHoldObj], QuaNoteList, HoldList):

    def _upcast(self, objList: List = None) -> QuaHoldList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: QuaHoldList
        """
        return QuaHoldList(objList)

    def data(self) -> List[QuaHoldObj]:
        return self



from __future__ import annotations
from PIL import Image, ImageDraw

from reamber.algorithms.analysis.playField.parts.PFDrawable import PFDrawable
from reamber.algorithms.analysis.playField import PlayField

class PFDrawNotes(PFDrawable):

    COL_DICT = [[0],
                [1,0],
                [0,2,0],
                [0,1,1,0],
                [0,1,2,1,0],
                [0,1,0,0,1,0],
                [0,1,0,2,0,1,0],
                [0,1,0,1,1,0,1,0],
                [0,1,0,1,2,1,0,1,0],
                [0,1,0,1,0,0,1,0,1,0],
                [0,1,0,1,0,2,0,1,0,1,0],
                [0,1,0,1,0,1,1,0,1,0,1,0],
                [0,1,0,1,0,1,2,1,0,1,0,1,0],
                [0,1,0,1,0,1,0,0,1,0,1,0,1,0],
                [0,1,0,1,0,1,0,2,0,1,0,1,0,1,0],
                [0,1,0,1,0,1,0,1,1,0,1,0,1,0,1,0],
                [0,1,0,1,0,1,0,1,2,1,0,1,0,1,0,1,0],
                [0,1,0,1,0,1,0,1,0,0,1,0,1,0,1,0,1,0]]

    def __init__(self,
                 img0OutlineColor: str = "#47fcff",
                 img0FillColor: str    = "#28b5b8",
                 img1OutlineColor: str = "#ffffff",
                 img1FillColor: str    = "#c2c2c2",
                 img2OutlineColor: str = "#f8ff30",
                 img2FillColor: str    = "#adb31e",
                 outlineWidth: int     = 2):
        """ The draws all the notes on the field

        The color used for each column is specified in the COL_DICT

        :param img0OutlineColor: The color to outline the first note image
        :param img0FillColor: The color to fill the first note image
        :param img1OutlineColor: The color to outline the second note image
        :param img1FillColor: The color to fill the second note image
        :param img2OutlineColor: The color to outline the third note image
        :param img2FillColor: The color to fill the third note image
        :param outlineWidth: The width of each outline. 0 for no outline
        """
        self.outlineWidth     = outlineWidth
        self.img0OutlineColor = img0OutlineColor
        self.img0FillColor    = img0FillColor
        self.img1OutlineColor = img1OutlineColor
        self.img1FillColor    = img1FillColor
        self.img2OutlineColor = img2OutlineColor
        self.img2FillColor    = img2FillColor

    def draw(self, pf: PlayField) -> PlayField:
        """ Refer to __init__ """

        imgs = [self._createNoteSet(pf=pf, fillColor=self.img0FillColor, outlineColor=self.img0OutlineColor,
                                    width=self.outlineWidth),
                self._createNoteSet(pf=pf, fillColor=self.img1FillColor, outlineColor=self.img1OutlineColor,
                                    width=self.outlineWidth),
                self._createNoteSet(pf=pf, fillColor=self.img2FillColor, outlineColor=self.img2OutlineColor,
                                    width=self.outlineWidth)]

        self._drawHits(imgs, pf)
        self._drawHolds(imgs, pf)

        return pf

    def _drawHits(self, imgs, pf: PlayField):
        for hit in pf.m.notes.hits():
            hitImg = imgs[self.COL_DICT[pf.keys - 1][hit.column]]['hit']
            pf.canvas.paste(hitImg,
                            (int(hit.column * (pf.noteWidth + pf.columnLineWidth)),
                            pf.canvasH - int((hit.offset - pf.start) / pf.durationPerPx) - pf.hitHeight),
                            hitImg)

    def _drawHolds(self, imgs, pf: PlayField):
        for hold in pf.m.notes.holds():
            holdHeadImg = imgs[self.COL_DICT[pf.keys - 1][hold.column]]['holdH']

            # DRAWS THE HEAD
            pf.canvas.paste(holdHeadImg,
                            (int(hold.column * (pf.noteWidth + pf.columnLineWidth)),
                            pf.canvasH - int((hold.offset - pf.start) / pf.durationPerPx) - pf.holdHeight),
                            holdHeadImg)

            holdTailImg = imgs[self.COL_DICT[pf.keys - 1][hold.column]]['holdT']

            # DRAWS THE TAIL
            pf.canvas.paste(holdTailImg,
                            (int(hold.column * (pf.noteWidth + pf.columnLineWidth)),
                            pf.canvasH - int((hold.tailOffset() - pf.start) / pf.durationPerPx) - pf.holdHeight),
                            holdTailImg)

            # DRAWS THE BODY
            holdImgHeight = int(hold.length / pf.durationPerPx) - pf.holdHeight + pf.HOLD_RESIZE_BUFFER

            # If too short we don't draw it
            if holdImgHeight > 0:
                holdImg = imgs[self.COL_DICT[pf.keys - 1][hold.column]]['holdB'].resize((pf.noteWidth, holdImgHeight))

                pf.canvas.paste(holdImg,
                                (int(hold.column * (pf.noteWidth + pf.columnLineWidth)),
                                pf.canvasH - int((hold.tailOffset() - pf.start) / pf.durationPerPx)),
                                holdImg)

    @staticmethod
    def _createHit(pf: PlayField, fillColor, outlineColor, width=4):
        img = Image.new(mode='RGBA', size=(pf.noteWidth, pf.hitHeight), color=(0, 0, 0, 0))
        draw = ImageDraw.Draw(img)
        path = [(0, 0),
                (0, pf.hitHeight - 1),
                (pf.noteWidth - 1, pf.hitHeight - 1),
                (pf.noteWidth - 1, 0),
                (0, 0)]
        draw.polygon(path, fill=fillColor)
        draw.line(path, fill=outlineColor, width=width)
        return img

    @staticmethod
    def _createHoldHead(pf: PlayField, fillColor, outlineColor, width=4):
        img = Image.new(mode='RGBA', size=(pf.noteWidth, pf.holdHeight), color=(0, 0, 0, 0))
        draw = ImageDraw.Draw(img)
        path = [(0, 0),
                (int(pf.noteWidth * 1 / 3), pf.holdHeight - int(width / 2)),
                (int(pf.noteWidth * 2 / 3), pf.holdHeight - int(width / 2)),
                (pf.noteWidth - 1, 0)]

        draw.polygon(path, fill=fillColor)
        draw.line(path, fill=outlineColor, width=width)
        return img

    @staticmethod
    def _createHoldBody(pf: PlayField, fillColor, outlineColor, width=4):
        img = Image.new(mode='RGBA', size=(pf.noteWidth, pf.holdHeight), color=(0, 0, 0, 0))
        draw = ImageDraw.Draw(img)
        draw.polygon([(0, 0),
                      (0, pf.noteWidth - 1),
                      (pf.noteWidth - 1, pf.hitHeight - 1),
                      (pf.noteWidth - 1, 0)],
                     fill=fillColor)
        draw.line([(0, 0),
                   (0, pf.holdHeight)],
                  fill=outlineColor, width=width)
        draw.line([(pf.noteWidth - 2, 0),
                   (pf.noteWidth - 2, pf.holdHeight)],
                  fill=outlineColor, width=width)
        return img

    @classmethod
    def _createHoldTail(cls, pf: PlayField, fillColor, outlineColor, width=4):
        """ It's just the inverted head """
        return cls._createHoldHead(pf, fillColor, outlineColor, width).transpose(Image.FLIP_TOP_BOTTOM)

    @classmethod
    def _createNoteSet(cls, pf: PlayField, fillColor, outlineColor, width=4):
        return {'hit': cls._createHit(pf, fillColor, outlineColor, width),
                'holdH': cls._createHoldHead(pf, fillColor, outlineColor, width),
                'holdB': cls._createHoldBody(pf, fillColor, outlineColor, width),
                'holdT': cls._createHoldTail(pf, fillColor, outlineColor, width)}



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj


class TestInit(unittest.TestCase):

    def testMixed(self):
        # Quick Init Mixed
        seq = SvSequence([100, (200, 2.0, True), (400, 3.0), SvObj(offset=800, multiplier=5.0, fixed=True)])
        self.assertEqual(len(seq), 4)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from reamber.algorithms.analysis.playField import PlayField
from reamber.algorithms.analysis.playField.parts.PFDrawable import PFDrawable


class PFDrawColumnLines(PFDrawable):

    def __init__(self,
                 color: str = "#2b2b2b"):
        """ Draws column lines, separating columns of notes

        :param color: The color of the column lines
        """
        self.color = color
        # The reason we don't isolate width here is because the canvas size, note placement is dependent on it
        # so, we make it a PlayField prop

    def draw(self, pf: PlayField) -> PlayField:
        """ Refer to __init__ """

        for colLine in range(1, pf.keys):  # Fencepost again, if key = 4, we draw on 1 2 3
            for w in range(pf.columnLineWidth):
                pf.canvasDraw.line([(colLine * pf.noteWidth + (colLine - 1) * pf.columnLineWidth + w, 0),
                                    (colLine * pf.noteWidth + (colLine - 1) * pf.columnLineWidth + w, pf.canvasH)],
                                    fill=self.color)
        return pf



from reamber.base.lists.notes.NoteList import NoteList
from typing import List, Type
from abc import ABC

from reamber.osu.OsuNoteObjMeta import OsuNoteObjMeta
from reamber.osu.OsuSampleSet import OsuSampleSet


class OsuNoteList(NoteList, ABC):
    def data(self) -> List[Type[OsuNoteObjMeta]]: pass

    def volumes(self) -> List[float]:
        return self.attribute('volume')

    def hitsoundFiles(self) -> List[str]:
        return self.attribute('hitsoundFile')

    def sampleSets(self) -> List[OsuSampleSet]:
        return self.attribute('sampleSet')

    def hitsoundSets(self) -> List[OsuSampleSet]:
        return self.attribute('hitsoundSet')

    def customSets(self) -> List[OsuSampleSet]:
        return self.attribute('customSet')

    def additionSets(self) -> List[OsuSampleSet]:
        return self.attribute('additionSet')



from reamber.algorithms.generate.sv.SvSequence import SvSequence
from reamber.algorithms.generate.sv.SvIO import SvIO
from reamber.algorithms.generate.sv.SvObj import SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


__all__ = ['SvObj', 'SvPkg', 'SvSequence', 'SvIO']



from dataclasses import dataclass
from typing import Dict

from reamber.base.HitObj import HitObj
from reamber.quaver.QuaNoteObjMeta import QuaNoteObjMeta


@dataclass
class QuaHitObj(QuaNoteObjMeta, HitObj):
    def asDict(self) -> Dict:
        """ Used to facilitate exporting as Qua from YAML """
        return {'StartTime': self.offset,
                'Lane': self.column + 1,
                'KeySounds': self.keySounds}



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMFakeObj import SMFakeObj
from typing import List


class SMFakeList(List[SMFakeObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMFakeList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMFakeList
        """
        return SMFakeList(objList)

    def data(self) -> List[SMFakeObj]:
        return self



from dataclasses import dataclass
from reamber.base.HoldObj import HoldObj
from reamber.o2jam.O2JNoteObjMeta import O2JNoteObjMeta


@dataclass
class O2JHoldObj(HoldObj, O2JNoteObjMeta):
    """ Defines the O2Jam Bpm Object

    The O2Jam Bpm Object is stored in binary file .ojn
    """
    pass



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.O2JToOsu import O2JToOsu
from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestOsuToQua(unittest.TestCase):

    # @profile
    def test_o2j1(self):
        # Complex BPM Points
        o2j = O2JMapSetObj()
        o2j.readFile(O2J_FLY_MAGPIE_OJN)

        osus = O2JToOsu.convert(o2j)
        osus[0].writeFile("out.osu")
        osus[1].writeFile("out.osu")
        osus[2].writeFile("out.osu")

    # @profile
    def test_o2j2(self):
        # Complex BPM Points
        o2j = O2JMapSetObj()
        o2j.readFile(O2J_CHECK_IT_OUT_OJN)

        osus = O2JToOsu.convert(o2j)
        osus[0].writeFile("out.osu")
        osus[1].writeFile("out.osu")
        osus[2].writeFile("out.osu")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import List
from reamber.base.lists.BpmList import BpmList
from reamber.quaver.QuaBpmObj import QuaBpmObj


class QuaBpmList(BpmList):

    def _upcast(self, objList: List = None) -> QuaBpmList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: QuaBpmList
        """
        return QuaBpmList(objList)

    def data(self) -> List[QuaBpmObj]:
        return self



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reamber",
    version="0.0.14",
    author="evening",
    author_email="dev_evening@hotmail.com",
    description="Vertical Scrolling Rhythm Game Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Eve-ning/reamber_base_py",
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        'pyyaml',
        'pandas',
        'matplotlib',
        'pillow'
    ]
)


from reamber.algorithms.generate.sv.SvSequence import SvSequence
from reamber.algorithms.generate.sv.SvPkg import SvPkg
from typing import Callable, List, Union
from numpy import arange


def svFuncSequencer(funcs: List[Union[float, Callable[[float], float], None]],
                    offsets: Union[List[float], float, None] = None,
                    repeats: int = 1,
                    repeatGap: float = 0,
                    startX: float = 0,
                    endX: float = 1
                    ):
    """ Sets up a sequence using functions.

    :param funcs: Funcs to generate values. \
        If List, values will be used directly. \
        If Callable, values will be called with the X. \
        If None, this will leave a gap in the sequence.
    :param offsets: Offsets to use on functions. \
        If List, offsets will be used to map the funcs. \
        If Float, all funcs are assumed to be separated by {float} ms. Starting from 0. \
        If None, all funcs are assumed to be separated by 1 ms. Starting from 0.
    :param repeats: The amount of repeats. This affects the increment of the X argument passed to the Callables. \
        If 0, only endX will be used.
    :param repeatGap: The gap between the repeats.
    :param startX: The starting X.
    :param endX: The ending X.
    """

    length = len(funcs)

    if offsets is None: offsets = list(range(0, length))
    # We use [:length] because sometimes arange will create too many for some reason (?)
    elif isinstance(offsets, (float, int)): offsets = list(arange(0, length * offsets, offsets))[:length]

    assert length == len(offsets)

    seq = SvSequence()

    for i, (offset, func) in enumerate(zip(offsets, funcs)):
        if isinstance(func, Callable): seq.appendInit([(offset, 0)])
        elif isinstance(func, (float, int)): seq.appendInit([(offset, func)])
        elif func is None: pass

    pkg = SvPkg.repeat(seq=seq, times=repeats, gap=repeatGap)

    nones = 0
    for funcI, func in enumerate(funcs):
        if func is None: nones += 1
        if isinstance(func, Callable):
            pkg.applyNth(func, funcI - nones, startX, endX)

    return pkg



from reamber.base.BpmObj import BpmObj
from reamber.base.HitObj import HitObj
from reamber.base.HoldObj import HoldObj
from reamber.base.MapObj import MapObj
from reamber.base.RAConst import RAConst
from reamber.base.TimedObj import TimedObj

__all__ = ['BpmObj', 'HitObj', 'HoldObj', 'MapObj', 'RAConst', 'TimedObj']



from __future__ import annotations

from typing import List, Type, TYPE_CHECKING
from abc import ABC, abstractmethod

from reamber.base.lists.TimedList import TimedList

if TYPE_CHECKING:
    from reamber.base.NoteObj import NoteObj


class NoteList(TimedList, ABC):
    """ Extends from the TimedList to give more base functions to NoteObjs
    """

    @abstractmethod
    def data(self) -> List[Type[NoteObj]]: pass

    def maxColumn(self) -> int:
        """ CALCULATES the key of the map
        Note that keys of the map isn't stored, it's dynamic and not a stored parameter.
        The function just finds the maximum column.
        """
        if len(self.columns()) == 0: return 0
        return max(self.columns())

    def columns(self) -> List[int]:
        return self.attribute('column')

    def inColumns(self, columns: List[int], inplace: bool = False) -> NoteList:
        """ Gets all objects that are in these columns """
        if inplace: self.__init__([obj for obj in self.data() if obj.column in columns])
        else: return self._upcast([obj for obj in self.data() if obj.column in columns])



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestSvSequence(unittest.TestCase):

    def test(self):
        # Quick Init Mixed
        seq = SvSequence([100, 200, 60, 400])
        seq.rescale(300, 800, inplace=True)


        pass


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from reamber.base.BpmObj import BpmObj


@dataclass
class O2JBpmObj(BpmObj):
    """ Defines the O2Jam Bpm Object

    The O2Jam Bpm Object is stored in binary file .ojn
    """
    pass



from __future__ import annotations
from reamber.o2jam.lists.notes.O2JNoteList import O2JNoteList
from reamber.o2jam.O2JHitObj import O2JHitObj
from typing import List


class O2JHitList(List[O2JHitObj], O2JNoteList):

    def _upcast(self, objList: List = None) -> O2JHitList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: O2JHitList
        """
        return O2JHitList(objList)

    def data(self) -> List[O2JHitObj]:
        return self



class RAConst :
    """ This class holds all constants that can be used throughout the program

    The class defines a helper classes to convert in units::

        hrToMin(10) == 600
        minToSec(10) == 600
        secToMSec(10) == 10000
        secToHr(60 * 60) == 1

    This also defines division classes, used in PlayField::

         {1: "#969696"
        , 2: "#b02323"
        , 3: "#8a23b0"
        , 4: "#2399b0"
        , 5: "#ffdfbd"
        , 6: "#b02378"
        , 8: "#b06c23"
        , 12: "#23b09d"
        , 16: "#3db023"
        , 24: "#94b023"
        , 32: "#23b052"}

    """

    HR_TO_MIN   : float = 60.0
    HR_TO_SEC   : float = 60.0 * 60.0
    HR_TO_MSEC  : float = 60.0 * 60.0 * 1000.0
    MIN_TO_HR   : float = 1.0 / 60.0
    MIN_TO_SEC  : float = 60.0
    MIN_TO_MSEC : float = 60.0 * 1000.0
    SEC_TO_HR   : float = 1.0 / (60.0 * 60.0)
    SEC_TO_MIN  : float = 1.0 / 60.0
    SEC_TO_MSEC : float = 1000.0
    MSEC_TO_HR  : float = 1.0 / (60.0 * 60.0 * 1000.0)
    MSEC_TO_MIN : float = 1.0 / (60.0 * 1000.0)
    MSEC_TO_SEC : float = 1.0 / 1000.0

    @staticmethod
    def hrToMin(hours): return float(hours * RAConst.HR_TO_MIN)
    @staticmethod
    def hrToSec(hours): return float(hours * RAConst.HR_TO_SEC)
    @staticmethod
    def hrToMSec(hours): return float(hours * RAConst.HR_TO_MSEC)
    @staticmethod
    def minToHr(mins): return float(mins * RAConst.MIN_TO_HR)
    @staticmethod
    def minToSec(mins): return float(mins * RAConst.MIN_TO_SEC)
    @staticmethod
    def minToMSec(mins): return float(mins * RAConst.MIN_TO_MSEC)
    @staticmethod
    def secToHr(secs): return float(secs * RAConst.SEC_TO_HR)
    @staticmethod
    def secToMin(secs): return float(secs * RAConst.SEC_TO_MIN)
    @staticmethod
    def secToMSec(secs): return float(secs * RAConst.SEC_TO_MSEC)
    @staticmethod
    def mSecToHr(msecs): return float(msecs * RAConst.MSEC_TO_HR)
    @staticmethod
    def mSecToMin(msecs): return float(msecs * RAConst.MSEC_TO_MIN)
    @staticmethod
    def mSecToSec(msecs): return float(msecs * RAConst.MSEC_TO_SEC)

    """ These are the colors to indicate snapping. Used for PlayField """
    DIVISION_COLORS = {1: "#969696"
                     , 2: "#b02323"
                     , 3: "#8a23b0"
                     , 4: "#2399b0"
                     , 5: "#ffdfbd"
                     , 6: "#b02378"
                     , 8: "#b06c23"
                     , 12: "#23b09d"
                     , 16: "#3db023"
                     , 24: "#94b023"
                     , 32: "#23b052"}



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestCopyTo(unittest.TestCase):
 
    def testCopy(self):
        # Test Copy
        seq = SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)])
        seqCopy = SvPkg.copyTo(seq=seq,offsets=[100, 200, 300]).combine(combineMethod=SvPkg.CombineMethod.DROP_BY_POINT)
        self.assertEqual(len(seqCopy), 4)


if __name__ == '__main__':
    unittest.main()



from reamber.osu.OsuMapObj import OsuMapObj, OsuSvObj, OsuBpmObj
from reamber.sm.SMMapSetObj import SMMapObj, SMBpmObj
from reamber.o2jam.O2JMapObj import O2JMapObj, O2JBpmObj
from reamber.quaver.QuaMapObj import QuaMapObj, QuaSvObj, QuaBpmObj
from reamber.algorithms.analysis.bpm.scrollSpeed import scrollSpeed
from reamber.algorithms.generate.sv.SvObj import SvObj
from typing import Union, List, Type, overload
from abc import abstractmethod, ABC


class SvIO(ABC):
    """ SV IO handles the input output functions for SvSequence.

    We don't integrate it like in PlayField because SvSequence can also be initialized very quickly with primitives.

    """

    @abstractmethod
    def __init__(self, _): ...
    @abstractmethod
    def data(self) -> List[SvObj]: ...

    @overload
    def writeAsSv(self, singularType: Type[OsuSvObj], **kwargs) -> List: ...
    @overload
    def writeAsSv(self, singularType: Type[QuaSvObj], **kwargs) -> List: ...
    def writeAsSv(self, singularType: Type, **kwargs) -> List:
        """ Writes the sequence as a List[singularType(offset=sv.offset, multiplier=sv.multiplier)]

        Must be able to accept 'offset' and 'multiplier' as argument.

        Allows **kwargs to input in every singularType during writing.

        Example::

            seq.writeAsSv(OsuSvObj, volume=20, kiai=True)

        :param singularType: A Type to specify, recommended to follow the overloaded Types."""
        return [singularType(offset=sv.offset, multiplier=sv.multiplier, **kwargs) for sv in self.data()]

    @overload
    def writeAsBpm(self, singularType: Type[OsuBpmObj], multiplication: float = 1.0, **kwargs) -> List: ...
    @overload
    def writeAsBpm(self, singularType: Type[QuaBpmObj], multiplication: float = 1.0, **kwargs) -> List: ...
    @overload
    def writeAsBpm(self, singularType: Type[SMBpmObj], multiplication: float = 1.0, **kwargs) -> List: ...
    @overload
    def writeAsBpm(self, singularType: Type[O2JBpmObj], multiplication: float = 1.0, **kwargs) -> List: ...
    def writeAsBpm(self, singularType: Type, multiplication: float = 1.0, **kwargs) -> List:
        """ Writes the sequence as a List[singularType(offset=sv.offset, bpm=sv.multiplier)]

        Must be able to accept 'offset' and 'bpm' as argument.

        Allows **kwargs to input in every singularType during writing.

        Multiplication multiplies the multiplier before exporting.

        Example::

            seq.writeAsBpm(OsuBpmObj, volume=20, kiai=True)

        :param singularType: A Type to specify, recommended to follow the overloaded Types.
        :param multiplication: The value to multiply before exporting the sv as a BPM.
        """
        return [singularType(offset=sv.offset, bpm=sv.multiplier * multiplication, **kwargs) for sv in self.data()]

    def readSvFromMap(self,
                      m: Union[OsuMapObj, QuaMapObj]):
        """ Reads the scroll velocities from maps.

        Inplace operation, doesn't return anything

        Only some map types have scroll velocities

        :param m: The map to read
        """
        self.__init__([(sv.offset, sv.multiplier) for sv in m.svs])

    def readTrueSvFromMap(self,
                          m: Union[OsuMapObj, O2JMapObj, SMMapObj, QuaMapObj],
                          centerBpm: float = None):
        """ Reads the true Scroll Velocity. That is, if present, SVs will be multiplied by the BPM.

        i.e. 200 Bpm and 2.0x SV -> 400.0x SV

        Inplace operation, doesn't return anything.

        :param m: The map to read
        :param centerBpm: The bpm to zero calculations on. If None, it'll just be the multiplication of bpm and sv.
        """
        self.__init__([(scroll['offset'], scroll['speed']) for scroll in scrollSpeed(m, centerBpm)])



from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.base.BpmObj import BpmObj
from reamber.sm.SMMapObjMeta import SMMapObjChartTypes
from reamber.sm.SMHitObj import SMHitObj
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.sm.SMBpmObj import SMBpmObj
from reamber.sm.lists.SMNotePkg import SMNotePkg
from reamber.sm.lists.SMBpmList import SMBpmList
from reamber.sm.lists.notes.SMHitList import SMHitList
from reamber.sm.lists.notes.SMHoldList import SMHoldList
from typing import List


class OsuToSM:
    @staticmethod
    def convert(osu: OsuMapObj) -> SMMapSetObj:
        """ Converts Osu to a SMMapset Obj

        Note that each osu map object will create a separate mapset, they are not merged

        :param osu: Osu Map
        :return: SM Mapset
        """

        # I haven't tested with non 4 keys, so it might explode :(

        assert osu.circleSize == 4

        hits: List[SMHitObj] = []
        holds: List[SMHoldObj] = []

        for hit in osu.notes.hits():
            hits.append(SMHitObj(offset=hit.offset, column=hit.column))
        for hold in osu.notes.holds():
            holds.append(SMHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

        bpms: List[BpmObj] = []

        for bpm in osu.bpms:
            bpms.append(SMBpmObj(offset=bpm.offset, bpm=bpm.bpm))

        smSet: SMMapSetObj = SMMapSetObj(
            music=osu.audioFileName,
            title=osu.title,
            titleTranslit=osu.titleUnicode,
            artist=osu.artist,
            artistTranslit=osu.artistUnicode,
            credit=osu.creator,
            background=osu.backgroundFileName,
            sampleStart=osu.previewTime,
            sampleLength=10,
            offset=osu.notes.firstOffset(),
            maps=[
                SMMapObj(
                    chartType=SMMapObjChartTypes.DANCE_SINGLE,
                    notes=SMNotePkg(hits=SMHitList(hits),
                                    holds=SMHoldList(holds)),
                    bpms=SMBpmList(bpms)
                )
            ]
        )

        return smSet



from __future__ import annotations
from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.base.BpmObj import BpmObj
from dataclasses import dataclass

MAX_BPM = 1e07
MIN_BPM = 1e-07

@dataclass
class OsuBpmObj(OsuTimingPointMeta, BpmObj):
    @staticmethod
    def codeToValue(code: float) -> float:
        """ Converts the data in the .osu file to the actual Bpm Value """
        return 60000.0 / code

    @staticmethod
    def valueToCode(value: float) -> float:
        """ Converts the actual Bpm Value to a writable float in .osu """
        return 60000.0 / value

    @staticmethod
    def readString(s: str) -> OsuBpmObj or None:
        """ Reads a single line under the [TimingPoints] Label. This must explicitly be a BPM Point. """
        if s.isspace(): return None

        sComma = s.split(",")
        if len(sComma) < 8: return None

        this = OsuBpmObj()
        assert sComma[6] == '1', "Unexpected SV Object in OsuBpmObj."
        this.offset = float(sComma[0])
        this.bpm = OsuBpmObj.codeToValue(float(sComma[1]))
        this.metronome = int(sComma[2])
        this.sampleSet = int(sComma[3])
        this.sampleSetIndex = int(sComma[4])
        this.volume = int(sComma[5])
        this.kiai = int(sComma[7])

        return this

    def writeString(self) -> str:
        """ Exports a .osu writable string """
        return f"{self.offset},{self.valueToCode(self.bpm)}," \
               f"{self.metronome},{self.sampleSet}," \
               f"{self.sampleSetIndex},{self.volume},{1},{int(self.kiai)}"



import unittest
from tests.test.RSC_PATHS import *

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.algorithms.analysis.bpm.scrollSpeed import scrollSpeed
import matplotlib.pyplot as plt
import pandas as pd

class TestScrollSpeed(unittest.TestCase):

    # @profile
    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_CARAVAN)
        pd.DataFrame(scrollSpeed(m)).set_index('offset').plot()
        plt.savefig("osu.png")

    # @profile
    def test_qua(self):
        m = QuaMapObj()
        m.readFile(QUA_NEURO_CLOUD)
        pd.DataFrame(scrollSpeed(m)).set_index('offset').plot()
        plt.savefig("qua.png")

    # @profile
    def test_sm(self):
        s = SMMapSetObj()
        s.readFile(SM_CARAVAN)
        pd.DataFrame(scrollSpeed(s.maps[0])).set_index('offset').plot()
        plt.savefig("sm.png")

    # @profile
    def test_o2j(self):
        s = O2JMapSetObj()
        s.readFile(O2J_FLY_MAGPIE_OJN)
        pd.DataFrame(scrollSpeed(s.maps[2])).set_index('offset').plot()
        plt.savefig("o2j.png")


if __name__ == '__main__':
    unittest.main()



from reamber.quaver.QuaMapObjMeta import QuaMapObjMeta
from reamber.base.MapObj import MapObj
from reamber.base.lists.TimedList import TimedList
from reamber.quaver.QuaSvObj import QuaSvObj
from reamber.quaver.QuaBpmObj import QuaBpmObj
from reamber.quaver.QuaHitObj import QuaHitObj
from reamber.quaver.QuaHoldObj import QuaHoldObj
from dataclasses import dataclass, field
from typing import List, Dict, Union
import yaml
from yaml import CLoader as Loader, CDumper as Dumper

from reamber.quaver.lists.QuaNotePkg import QuaNotePkg
from reamber.quaver.lists.QuaBpmList import QuaBpmList
from reamber.quaver.lists.QuaSvList import QuaSvList


@dataclass
class QuaMapObj(QuaMapObjMeta, MapObj):

    notes: QuaNotePkg = field(default_factory=lambda: QuaNotePkg())
    bpms:  QuaBpmList = field(default_factory=lambda: QuaBpmList())
    svs:   QuaSvList  = field(default_factory=lambda: QuaSvList())

    def data(self) -> Dict[str, TimedList]:
        """ Gets the notes, bpms and svs as a dictionary """
        return {'notes': self.notes,
                'bpms': self.bpms,
                'svs': self.svs}

    def readFile(self, filePath: str):
        """ Reads a .qua, loads inplace, hence it doesn't return anything

        :param filePath: The path to the .qua file."""

        self.__init__()

        with open(filePath, "r", encoding="utf8") as f:
            # Reading with CReader is much faster
            file = yaml.load(f, Loader=Loader)
        # We pop them so as to reduce the size needed to pass to _readMeta
        self._readNotes(file.pop('HitObjects'))
        self._readBpms(file.pop('TimingPoints'))
        self._readSVs(file.pop('SliderVelocities'))
        self._readMetadata(file)

    def writeFile(self, filePath: str):
        """ Writes a .qua, doesn't return anything.

        :param filePath: The path to a new .qua file."""
        file = self._writeMeta()

        bpm: QuaBpmObj
        file['TimingPoints'] = [bpm.asDict() for bpm in self.bpms]
        sv: QuaSvObj
        file['SliderVelocities'] = [sv.asDict() for sv in self.svs]
        note: Union[QuaHitObj, QuaHoldObj]

        # This long comprehension squishes the hits: {} and holds: {} to a list for asDict operation
        # noinspection PyTypeChecker
        file['HitObjects'] = [i.asDict() for j in [v for k, v in self.notes.data().items()] for i in j]
        with open(filePath, "w+", encoding="utf8") as f:
            # Writing with CDumper is much faster
            f.write(yaml.dump(file, default_flow_style=False, sort_keys=False, Dumper=Dumper))

    def _readBpms(self, bpms: List[Dict]):
        for bpm in bpms:
            self.bpms.append(QuaBpmObj(offset=bpm['StartTime'], bpm=bpm['Bpm']))

    def _readSVs(self, svs: List[Dict]):
        for sv in svs:
            self.svs.append(QuaSvObj(offset=sv['StartTime'], multiplier=sv['Multiplier']))

    def _readNotes(self, notes: List[Dict]):
        for note in notes:
            offset = note['StartTime']
            column = note['Lane'] - 1
            keySounds = note['KeySounds']
            if "EndTime" in note.keys():
                self.notes.holds().append(QuaHoldObj(offset=offset, length=note['EndTime'] - offset,
                                                      column=column, keySounds=keySounds))
            else:
                self.notes.hits().append(QuaHitObj(offset=offset, column=column, keySounds=keySounds))



import unittest
from reamber.algorithms.generate.sv.generators.svNormalizeBpm import svNormalizeBpm
from reamber.base.lists.BpmList import BpmList, BpmObj


class TestNormalize(unittest.TestCase):

    def testNormalize(self):
        seq = svNormalizeBpm(BpmList([BpmObj(0, 200), BpmObj(100, 50), BpmObj(300, 100)]), 100)
        self.assertAlmostEqual(seq[0].multiplier, 0.5)
        self.assertAlmostEqual(seq[1].multiplier, 2.0)
        self.assertAlmostEqual(seq[2].multiplier, 1.0)

if __name__ == '__main__':
    unittest.main()
